# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
========================================
Signals (:mod:`qiskit_dynamics.signals`)
========================================

.. currentmodule:: qiskit_dynamics.signals

This module contains classes for representing the time-dependent coefficients in matrix differential
equations.

These classes, referred to as *signals*, represent classes of real-valued functions, either of the
form, or built from functions of the following form:

.. math::
    s(t) = \textnormal{Re}[f(t)e^{i(2 \pi \nu t + \phi)}],

where

    * :math:`f` is a complex-valued function called the *envelope*,
    * :math:`\nu \in \mathbb{R}` is the *carrier frequency*, and
    * :math:`\phi \in \mathbb{R}` is the *phase*.

Furthermore, this module contains *transfer functions* which transform one or more signal into other
signals.

Signal API summary
==================

All signal classes share a common API for evaluation and visualization:

    * The signal value at a given time ``t`` is evaluated by treating the ``signal`` as a callable:
      ``signal(t)``.
    * The envelope :math:`f(t)` is evaluated via: ``signal.envelope(t)``.
    * The complex value :math:`f(t)e^{i(2 \pi \nu t + \phi)}` via: ``signal.complex_value(t)``.
    * The ``signal.draw`` method provides a common visualization interface.

In addition to the above, all signal types allow for algebraic operations, which should be
understood in terms of algebraic operations on functions. E.g. two signals can be added together via

.. code-block:: python

    signal_sum = signal1 + signal2

and satisfy

.. code-block:: python

    signal_sum(t) == signal1(t) + signal2(t)

Signal multiplication is defined similarly, and signals can be added or multiplied with constants as
well.

The remainder of this document gives further detail about some special functionality of these
classes, but the following table provides a list of the different signal classes, along with a high
level description of their role.

.. list-table:: Types of signal objects
   :widths: 10 50
   :header-rows: 1

   * - Class name
     - Description
   * - :class:`~qiskit_dynamics.signals.Signal`
     - Envelope specified as a python ``Callable``, allowing for complete generality.
   * - :class:`~qiskit_dynamics.signals.DiscreteSignal`
     - Piecewise constant envelope, implemented with array-based operations, geared towards
       performance.
   * - :class:`~qiskit_dynamics.signals.SignalSum`
     - A sum of :class:`~qiskit_dynamics.signals.Signal` or
       :class:`~qiskit_dynamics.signals.DiscreteSignal` objects. Evaluation of envelopes returns an
       array of envelopes in the sum.
   * - :class:`~qiskit_dynamics.signals.DiscreteSignalSum`
     - A sum of :class:`~qiskit_dynamics.signals.DiscreteSignal` objects with the same start time,
       number of samples, and sample duration. Implemented with array-based operations.


Constant Signal
===============

:class:`~qiskit_dynamics.signals.Signal` supports specification of a *constant signal*:

.. code-block:: python

    const = Signal(2.)

This initializes the object to always return the constant ``2.``, and allows constants to be treated
on the same footing as arbitrary :class:`~qiskit_dynamics.signals.Signal` instances. A
:class:`~qiskit_dynamics.signals.Signal` operating in constant-mode can be checked via the boolean
attribute ``const.is_constant``.


Algebraic operations
====================

Algebraic operations are supported by the :class:`~qiskit_dynamics.signals.SignalSum` object. Any
two signal classes can be added together, producing a :class:`~qiskit_dynamics.signals.SignalSum`.
Multiplication is also supported via :class:`~qiskit_dynamics.signals.SignalSum` using the identity:

.. math::

    Re[f(t)e^{i(2 \pi \nu t + \phi)}] \times &Re[g(t)e^{i(2 \pi \omega t + \psi)}]
         \\&= Re[\frac{1}{2} f(t)g(t)e^{i(2\pi (\omega + \nu)t + (\phi + \psi))} ]
          + Re[\frac{1}{2} f(t)\overline{g(t)}e^{i(2\pi (\omega - \nu)t + (\phi - \psi))} ].

I.e. multiplication of two base signals produces a :class:`~qiskit_dynamics.signals.SignalSum` with
two elements, whose envelopes, frequencies, and phases are as given by the above formula.
Multiplication of sums is handled via distribution of this formula over the sum.

In the special case that :class:`~qiskit_dynamics.signals.DiscreteSignal`\s with compatible sample
structure (same number of samples, ``dt``, and start time) are added together, a
:class:`~qiskit_dynamics.signals.DiscreteSignalSum` is produced.
:class:`~qiskit_dynamics.signals.DiscreteSignalSum` stores a sum of compatible
:class:`~qiskit_dynamics.signals.DiscreteSignal`\s by joining the underlying arrays, so that the sum
can be evaluated using purely array-based operations. Multiplication of
:class:`~qiskit_dynamics.signals.DiscreteSignal`\s with compatible sample structure is handled
similarly.

Sampling
========

Both :class:`~qiskit_dynamics.signals.DiscreteSignal` and
:class:`~qiskit_dynamics.signals.DiscreteSignalSum` feature constructors
(:meth:`~qiskit_dynamics.signals.DiscreteSignal.from_Signal` and
:meth:`~qiskit_dynamics.signals.DiscreteSignalSum.from_SignalSum` respectively) which build an
instance by sampling a :class:`~qiskit_dynamics.signals.Signal` or
:class:`~qiskit_dynamics.signals.SignalSum`. These constructors have the option to just sample the
envelope (and keep the carrier analog), or to also sample the carrier. Below is a visualization of a
signal superimposed with sampled versions, both in the case of sampling the carrier, and in the case
of sampling just the envelope (and keeping the carrier analog).

.. jupyter-execute::
    :hide-code:

    from qiskit_dynamics.signals import Signal, DiscreteSignal
    from matplotlib import pyplot as plt

    # discretize a signal with and without samplying the carrier
    signal = Signal(lambda t: t, carrier_freq=2.)
    discrete_signal = DiscreteSignal.from_Signal(
        signal, dt=0.1, start_time=0., n_samples=10, sample_carrier=True
    )
    discrete_signal2 = DiscreteSignal.from_Signal(signal, dt=0.1, start_time=0., n_samples=10)

    # plot the signal against each discretization
    fig, axs = plt.subplots(1, 2, figsize=(14, 4))
    signal.draw(t0=0., tf=1., n=100, axis=axs[0])
    discrete_signal.draw(
        t0=0., tf=1., n=100, axis=axs[0], title='Signal v.s. Sampled envelope and carrier'
    )
    signal.draw(t0=0., tf=1., n=100, axis=axs[1])
    discrete_signal2.draw(
        t0=0., tf=1., n=100, axis=axs[1], title='Signal v.s. Sampled envelope'
    )


Transfer Functions
==================

A transfer function is a mapping from one or more :class:`Signal` to one or more :class:`Signal`.
Transfer functions can, for example, be used to model the effect of the electronics finite
response. The code below shows the example of an :class:`IQMixer`. Here, two signals modulated at
100 MHz and with a relative :math:`\pi/2` phase shift are passed through an IQ-mixer with a
carrier frequency of 400 MHz to create a signal at 500 MHz. Note that the code below does not make
any assumptions about the time and frequency units which we interpret as ns and GHz, respectively.

.. jupyter-execute::

    import numpy as np
    from qiskit_dynamics.signals import DiscreteSignal, Sampler, IQMixer

    dt = 0.25
    in_phase = DiscreteSignal(dt, [1.0]*200, carrier_freq=0.1, phase=0)
    quadrature = DiscreteSignal(dt, [1.0]*200, carrier_freq=0.1, phase=np.pi/2)

    sampler = Sampler(dt/25, 5000)
    in_phase = sampler(in_phase)
    quadrature = sampler(quadrature)

    mixer = IQMixer(0.4)
    rf = mixer(in_phase, quadrature)

    fig, axs = plt.subplots(1, 2, figsize=(14, 4))
    in_phase.draw(0, 25, 100, axis=axs[0])
    quadrature.draw(0, 25, 100, axis=axs[0], title='In-phase and quadrature signals')
    rf.draw(0, 24, 2000, axis=axs[1], title='Mixer output')

Signal Classes
==============

.. autosummary::
   :toctree: ../stubs/

   Signal
   DiscreteSignal
   SignalSum
   DiscreteSignalSum
   SignalList

Transfer Function Classes
=========================

.. autosummary::
   :toctree: ../stubs/

   Convolution
   IQMixer
"""

from .signals import Signal, DiscreteSignal, SignalSum, DiscreteSignalSum, SignalList
from .transfer_functions import Convolution, Sampler, IQMixer



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name,no-member,attribute-defined-outside-init

r"""
Solver functions.
"""

from typing import Optional, Union, Callable, Tuple, TypeVar
from warnings import warn

from scipy.integrate import OdeSolver

from scipy.integrate._ivp.ivp import OdeResult

from qiskit import QiskitError
from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias import ArrayLike

from qiskit_dynamics.models import (
    BaseGeneratorModel,
    GeneratorModel,
    LindbladModel,
)
from qiskit_dynamics.models.hamiltonian_model import HamiltonianModel

from .solver_utils import is_lindblad_model_not_vectorized
from .fixed_step_solvers import (
    RK4_solver,
    jax_RK4_solver,
    scipy_expm_solver,
    lanczos_diag_solver,
    jax_lanczos_diag_solver,
    jax_expm_solver,
    jax_RK4_parallel_solver,
    jax_expm_parallel_solver,
)
from .scipy_solve_ivp import scipy_solve_ivp, SOLVE_IVP_METHODS
from .jax_odeint import jax_odeint
from .diffrax_solver import diffrax_solver

ODE_METHODS = (
    ["RK45", "RK23", "BDF", "DOP853", "Radau", "LSODA"]  # scipy solvers
    + ["RK4"]  # fixed step solvers
    + ["jax_odeint", "jax_RK4"]  # jax solvers
)
LMDE_METHODS = [
    "scipy_expm",
    "lanczos_diag",
    "jax_lanczos_diag",
    "jax_expm",
    "jax_expm_parallel",
    "jax_RK4_parallel",
]

# diffrax solver type placeholder
# pylint: disable=typevar-name-mismatch
DiffraxAbstractSolver = TypeVar("AbstractSolver")


def _is_jax_method(method: any) -> bool:
    """Check if method is a jax solver method."""
    if method in [
        "jax_odeint",
        "jax_RK4",
        "jax_expm",
        "jax_expm_parallel",
        "jax_RK4_parallel",
        "jax_lanczos_diag",
    ]:
        return True

    # only other jax methods are diffrax methods
    return _is_diffrax_method(method)


def _is_diffrax_method(method: any) -> bool:
    """Check if method is a diffrax method."""
    try:
        from diffrax import AbstractSolver

        return isinstance(method, AbstractSolver)
    except ImportError:
        return False


def _lanczos_validation(
    rhs: Union[Callable, BaseGeneratorModel],
    t_span: ArrayLike,
    y0: ArrayLike,
    k_dim: int,
):
    """Validation checks to run lanczos based solvers."""
    t_span = unp.asarray(t_span)
    y0 = unp.asarray(y0)
    if isinstance(rhs, BaseGeneratorModel):
        if not isinstance(rhs, HamiltonianModel):
            raise QiskitError(
                """Lanczos solver can only be used for HamiltonianModel or function-based
                    anti-Hermitian generators."""
            )
        if "sparse" not in rhs.array_library:
            warn(
                """lanczos_diag should be used with a generator in sparse mode
                for better performance.""",
                category=Warning,
                stacklevel=2,
            )

    dim = rhs(t_span[0]).shape[0]
    if k_dim > dim:
        raise QiskitError("k_dim can be no larger than the dimension of the generator.")

    if y0.ndim not in [1, 2]:
        raise QiskitError("y0 must be 1d or 2d.")


def solve_ode(
    rhs: Union[Callable, BaseGeneratorModel],
    t_span: ArrayLike,
    y0: ArrayLike,
    method: Optional[Union[str, OdeSolver, DiffraxAbstractSolver]] = "DOP853",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
) -> OdeResult:
    r"""General interface for solving Ordinary Differential Equations (ODEs).

    ODEs are differential equations of the form

    .. math::

        \dot{y}(t) = f(t, y(t)),

    where :math:`f` is a callable function and the state :math:`y(t)` is an
    arbitrarily-shaped complex array.

    The ``method`` argument exposes a variety of underlying ODE solvers. Optional
    arguments for any of the solver routines can be passed via ``kwargs``.
    Available methods are:

    - ``scipy.integrate.solve_ivp`` - supports methods
      ``['RK45', 'RK23', 'BDF', 'DOP853', 'Radau', 'LSODA']`` or by passing a valid
      ``scipy`` :class:`OdeSolver` instance.
    - ``'RK4'``: A fixed-step 4th order Runge-Kutta solver.
      Requires additional kwarg ``max_dt``, indicating the maximum step
      size to take. This solver will break integration periods into even
      sub-intervals no larger than ``max_dt``, and step over each sub-interval
      using the standard 4th order Runge-Kutta integration rule.
    - ``'jax_RK4'``: JAX backend implementation of ``'RK4'`` method.
    - ``'jax_odeint'``: Calls ``jax.experimental.ode.odeint`` variable step solver.
    - ``diffrax.diffeqsolve`` - a JAX solver function, called by passing ``method``
      as a valid ``diffrax.AbstractSolver`` instance. Requires the ``diffrax`` library.

    Results are returned as a :class:`OdeResult` object.

    Args:
        rhs: RHS function :math:`f(t, y)`.
        t_span: ``Tuple`` or ``list`` of initial and final time.
        y0: State at initial time.
        method: Solving method to use.
        t_eval: Times at which to return the solution. Must lie within ``t_span``. If unspecified,
                the solution will be returned at the points in ``t_span``.
        **kwargs: Additional arguments to pass to the solver.

    Returns:
        OdeResult: Results object.

    Raises:
        QiskitError: If specified method does not exist.
    """

    if method not in ODE_METHODS and not (
        (isinstance(method, type) and (issubclass(method, OdeSolver))) or _is_diffrax_method(method)
    ):
        raise QiskitError("Method " + str(method) + " not supported by solve_ode.")

    y0 = unp.asarray(y0)

    if isinstance(rhs, BaseGeneratorModel):
        _, solver_rhs, y0, model_in_frame_basis = setup_generator_model_rhs_y0_in_frame_basis(
            rhs, y0
        )
    else:
        solver_rhs = rhs

    # solve the problem using specified method
    if method in SOLVE_IVP_METHODS or (isinstance(method, type) and issubclass(method, OdeSolver)):
        results = scipy_solve_ivp(solver_rhs, t_span, y0, method, t_eval=t_eval, **kwargs)
    elif isinstance(method, str) and method == "RK4":
        results = RK4_solver(solver_rhs, t_span, y0, t_eval=t_eval, **kwargs)
    elif isinstance(method, str) and method == "jax_RK4":
        results = jax_RK4_solver(solver_rhs, t_span, y0, t_eval=t_eval, **kwargs)
    elif isinstance(method, str) and method == "jax_odeint":
        results = jax_odeint(solver_rhs, t_span, y0, t_eval=t_eval, **kwargs)
    elif _is_diffrax_method(method):
        results = diffrax_solver(solver_rhs, t_span, y0, method=method, t_eval=t_eval, **kwargs)

    # convert results out of frame basis if necessary
    if isinstance(rhs, BaseGeneratorModel):
        if not model_in_frame_basis:
            results.y = results_y_out_of_frame_basis(rhs, results.y, y0.ndim)

        # convert model back to original basis
        rhs.in_frame_basis = model_in_frame_basis

    return results


def solve_lmde(
    generator: Union[Callable, BaseGeneratorModel],
    t_span: ArrayLike,
    y0: ArrayLike,
    method: Optional[Union[str, OdeSolver, DiffraxAbstractSolver]] = "DOP853",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
) -> OdeResult:
    r"""General interface for solving Linear Matrix Differential Equations (LMDEs)
    in standard form.

    LMDEs in standard form are differential equations of the form:

    .. math::

        \dot{y}(t) = G(t)y(t).

    where :math:`G(t)` is a square matrix valued-function called the *generator*, and :math:`y(t)`
    is an array of appropriate shape.

    Thus function accepts :math:`G(t)` as a ``qiskit_dynamics`` model class, or as an arbitrary
    callable.

    .. note::

        Not all model classes are by-default in standard form. E.g.
        :class:`~qiskit_dynamics.models.LindbladModel` represents an LMDE which is not typically
        written in standard form. As such, using LMDE-specific methods with this generator requires
        the equation to be vectorized.

    The ``method`` argument exposes solvers specialized to both LMDEs, as well as general ODE
    solvers. If the method is not specific to LMDEs, the problem will be passed to
    :meth:`~qiskit_dynamics.solve_ode` by automatically setting up the RHS function :math:`f(t, y) =
    G(t)y`.

    Optional arguments for any of the solver routines can be passed via ``kwargs``. Available
    LMDE-specific methods are:

    - ``'scipy_expm'``: A fixed-step matrix-exponential solver using ``scipy.linalg.expm``. Requires
      additional kwarg ``max_dt`` indicating the maximum step size to take. This solver will break
      integration periods into even sub-intervals no larger than ``max_dt`` and solve over each
      sub-interval. The optional kwarg ``magnus_order`` controls the integration rule: if
      ``magnus_order==1``, the generator is sampled at the interval midpoint and exponentiated, and
      if ``magnus_order==2`` or ``magnus_order==3``, higher-order exponentiation rules are adopted
      from :footcite:`blanes_magnus_2009`. The ``magnus_order`` parameter defaults to ``1``.
    - ``'lanczos_diag'``: A fixed-step matrix-exponential solver, similar to ``'scipy_expm'`` but
      restricted to anti-Hermitian generators. The matrix exponential is performed by diagonalizing
      an approximate projection of the generator to a small subspace (the Krylov Subspace), obtained
      via the Lanczos algorithm, and then exponentiating the eigenvalues. Requires additional kwargs
      ``max_dt`` and ``k_dim`` indicating the maximum step size to take and Krylov subspace
      dimension, respectively. ``k_dim`` acts as an adjustable accuracy parameter and can be no
      larger than the dimension of the generator. The method is recommended for sparse systems with
      large dimension.
    - ``'jax_lanczos_diag'``: JAX implementation of ``'lanczos_diag'``, with the same arguments and
      behaviour. Note that this method contains calls to ``jax.numpy.eigh``, which may have limited
      validity when automatically differentiated.
    - ``'jax_expm'``: JAX-implemented version of ``'scipy_expm'``, with the same arguments and
      behaviour. Note that this method cannot be used for a model using a sparse array library.
    - ``'jax_expm_parallel'``: Same as ``'jax_expm'``, however all loops are implemented using
      parallel operations. I.e. all matrix-exponentials for taking a single step are computed in
      parallel using ``jax.vmap``, and are subsequently multiplied together in parallel using
      ``jax.lax.associative_scan``. This method is only recommended for use with GPU execution. Note
      that this method cannot be used for a model using a sparse array library.
    - ``'jax_RK4_parallel'``: 4th order Runge-Kutta fixed step solver. Under the assumption of the
      structure of an LMDE, utilizes the same parallelization approach as ``'jax_expm_parallel'``,
      however the single step rule is the standard 4th order Runge-Kutta rule, rather than
      matrix-exponentiation. Requires and utilizes the ``max_dt`` kwarg in the same manner as
      ``method='scipy_expm'``. This method is only recommended for use with GPU execution.

    Results are returned as a :class:`OdeResult` object.

    Args:
        generator: Representation of generator function :math:`G(t)`.
        t_span: ``Tuple`` or `list` of initial and final time.
        y0: State at initial time.
        method: Solving method to use.
        t_eval: Times at which to return the solution. Must lie within ``t_span``. If unspecified,
            the solution will be returned at the points in ``t_span``.
        **kwargs: Additional arguments to pass to the solver.

    Returns:
        OdeResult: Results object.

    Raises:
        QiskitError: If specified method does not exist,
                     if dimension of ``y0`` is incompatible with generator dimension, or if an
                     LMDE-specific method is passed with a LindbladModel.
    Additional Information:
        While all :class:`~qiskit_dynamics.models.BaseGeneratorModel` subclasses represent LMDEs,
        they are not all in standard form by defualt. Using an LMDE-specific models like
        :class:`~qiskit_dynamics.models.LindbladModel` requires first setting the model to be
        vectorized.
    """

    # delegate to solve_ode if necessary
    if (
        method in ODE_METHODS
        or (isinstance(method, type) and (issubclass(method, OdeSolver)))
        or _is_diffrax_method(method)
    ):
        if isinstance(generator, BaseGeneratorModel):
            rhs = generator
        else:
            # treat generator as a function
            def rhs(t, y):
                return generator(t) @ y

        return solve_ode(rhs, t_span, y0, method=method, t_eval=t_eval, **kwargs)

    # raise error if neither an ODE_METHOD or an LMDE_METHOD
    if method not in LMDE_METHODS:
        raise QiskitError(f"Method {method} not supported by solve_lmde.")

    # lmde-specific methods can't be used with LindbladModel unless vectorized
    if is_lindblad_model_not_vectorized(generator):
        raise QiskitError(
            "LMDE-specific methods with LindbladModel requires setting a vectorized=True."
        )

    y0 = unp.asarray(y0)

    # setup generator and rhs functions to pass to numerical methods
    if isinstance(generator, BaseGeneratorModel):
        solver_generator, _, y0, model_in_frame_basis = setup_generator_model_rhs_y0_in_frame_basis(
            generator, y0
        )
    else:
        solver_generator = generator

    if method == "scipy_expm":
        results = scipy_expm_solver(solver_generator, t_span, y0, t_eval=t_eval, **kwargs)
    elif "lanczos_diag" in method:
        _lanczos_validation(generator, t_span, y0, kwargs["k_dim"])
        if method == "lanczos_diag":
            results = lanczos_diag_solver(solver_generator, t_span, y0, t_eval=t_eval, **kwargs)
        elif method == "jax_lanczos_diag":
            results = jax_lanczos_diag_solver(solver_generator, t_span, y0, t_eval=t_eval, **kwargs)
    elif method == "jax_expm":
        if isinstance(generator, BaseGeneratorModel) and "sparse" in generator.array_library:
            raise QiskitError("jax_expm cannot be used with a generator in sparse mode.")
        results = jax_expm_solver(solver_generator, t_span, y0, t_eval=t_eval, **kwargs)
    elif method == "jax_expm_parallel":
        results = jax_expm_parallel_solver(solver_generator, t_span, y0, t_eval=t_eval, **kwargs)
    elif method == "jax_RK4_parallel":
        results = jax_RK4_parallel_solver(solver_generator, t_span, y0, t_eval=t_eval, **kwargs)

    # convert results to correct basis if necessary
    if isinstance(generator, BaseGeneratorModel):
        if not model_in_frame_basis:
            # pylint: disable=possibly-used-before-assignment
            results.y = results_y_out_of_frame_basis(generator, results.y, y0.ndim)

        generator.in_frame_basis = model_in_frame_basis

    return results


def setup_generator_model_rhs_y0_in_frame_basis(
    generator_model: BaseGeneratorModel, y0: ArrayLike
) -> Tuple[Callable, Callable, ArrayLike]:
    """Helper function for setting up a subclass of
    :class:`~qiskit_dynamics.models.BaseGeneratorModel` to be solved in the frame basis.

    Note: this function modifies ``generator_model`` to function in the frame basis.

    Args:
        generator_model: Subclass of :class:`~qiskit_dynamics.models.BaseGeneratorModel`.
        y0: Initial state.

    Returns:
        Callable for generator in frame basis, callable for RHS in frame basis, y0
        in frame basis, and boolean indicating whether model was already specified in frame basis.
    """

    model_in_frame_basis = generator_model.in_frame_basis

    # if model not specified in frame basis, transform initial state into frame basis
    if not model_in_frame_basis:
        if isinstance(generator_model, LindbladModel) and generator_model.vectorized:
            if generator_model.rotating_frame.frame_basis is not None:
                y0 = generator_model.rotating_frame.vectorized_frame_basis_adjoint @ y0
        elif isinstance(generator_model, LindbladModel):
            y0 = generator_model.rotating_frame.operator_into_frame_basis(y0)
        elif isinstance(generator_model, GeneratorModel):
            y0 = generator_model.rotating_frame.state_into_frame_basis(y0)

    # set model to operator in frame basis
    generator_model.in_frame_basis = True

    # define rhs functions in frame basis
    def generator(t):
        return generator_model(t)

    def rhs(t, y):
        return generator_model(t, y)

    return generator, rhs, y0, model_in_frame_basis


def results_y_out_of_frame_basis(
    generator_model: BaseGeneratorModel, results_y: ArrayLike, y0_ndim: int
) -> ArrayLike:
    """Convert the results of a simulation for :class:`~qiskit_dynamics.models.BaseGeneratorModel`
    out of the frame basis.

    Args:
        generator_model: Subclass of :class:`~qiskit_dynamics.models.BaseGeneratorModel`.
        results_y: Array whose first index corresponds to the evaluation points of the state
                   for the results of ``solve_lmde`` or ``solve_ode``.
        y0_ndim: Number of dimensions of initial state.

    Returns:
        Callable for generator in frame basis, Callable for RHS in frame basis, and y0
        transformed to frame basis.
    """
    # for left multiplication cases, if number of input dimensions is 1
    # vectorized basis transformation requires transposing before and after
    if y0_ndim == 1:
        results_y = results_y.T

    if isinstance(generator_model, LindbladModel) and generator_model.vectorized:
        if generator_model.rotating_frame.frame_basis is not None:
            results_y = generator_model.rotating_frame.vectorized_frame_basis @ results_y
    elif isinstance(generator_model, LindbladModel):
        results_y = generator_model.rotating_frame.operator_out_of_frame_basis(results_y)
    else:
        results_y = generator_model.rotating_frame.state_out_of_frame_basis(results_y)

    if y0_ndim == 1:
        results_y = results_y.T

    return results_y



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Model class for perturbative expansion of solution of an LMDE.
"""

from typing import Optional, Tuple, Callable, List, Union

import numpy as np
from numpy.polynomial.chebyshev import chebpts1, chebvander, chebval

from multiset import Multiset

from qiskit import QiskitError
from qiskit.quantum_info import Operator

from qiskit_dynamics import ArrayLike
from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import Signal, RotatingFrame
from qiskit_dynamics.perturbation import solve_lmde_perturbation, ArrayPolynomial
from qiskit_dynamics.solvers.solver_functions import _is_diffrax_method, _is_jax_method

try:
    import jax.numpy as jnp
    from jax.lax import scan
except ImportError:
    pass


class ExpansionModel:
    """Represents a perturbative expansion of an LMDE."""

    def __init__(
        self,
        operators: List[Operator],
        rotating_frame: Union[ArrayLike, RotatingFrame, None],
        dt: float,
        carrier_freqs: ArrayLike,
        chebyshev_orders: List[int],
        expansion_method: Optional[str] = "dyson",
        expansion_order: Optional[int] = None,
        expansion_labels: Optional[List[Multiset]] = None,
        integration_method: Optional[str] = None,
        include_imag: Optional[List[bool]] = None,
        **kwargs,
    ):
        r"""Initialize.

        Args:
            operators: List of constant operators specifying the operators with signal coefficients.
            rotating_frame: Rotating frame to setup the solver in. Must be Hermitian or
                anti-Hermitian.
            dt: Fixed step size to compile to.
            carrier_freqs: Carrier frequencies of the signals in the generator decomposition.
            chebyshev_orders: Approximation degrees for each signal over the interval [0, dt].
            expansion_method: Either ``'dyson'`` or ``'magnus'``.
            expansion_order: Order of perturbation terms to compute up to. Specifying this argument
                results in computation of all terms up to the given order. Can be used in
                conjunction with ``expansion_terms``.
            expansion_labels: Specific perturbation terms to compute. If both ``expansion_order``
                and ``expansion_terms`` are specified, then all terms up to ``expansion_order`` are
                computed, along with the additional terms specified in ``expansion_terms``. Labels
                are specified either as ``Multiset`` or as valid arguments to the ``Multiset``
                constructor. This function further requires that ``Multiset``\s consist only of
                non-negative integers.
            integration_method: ODE solver method to use when computing perturbation terms.
            include_imag: List of bools determining whether to keep imaginary components in
                the signal approximation. Defaults to True for all signals.
            kwargs: Additional arguments to pass to the solver when computing perturbation terms.

        Raises:
            QiskitError: If invalid ``expansion_method`` passed, or if arguments have incompatible
                lengths.
        """

        if expansion_method not in ["dyson", "magnus"]:
            raise QiskitError("ExpansionModel only accepts expansion_method 'dyson' or 'magnus'.")

        if len(operators) != len(carrier_freqs):
            raise QiskitError("carrier_freqs must have the same length as operators.")

        if len(operators) != len(chebyshev_orders):
            raise QiskitError("chebyshev_orders must have the same length as operators.")

        self._expansion_method = expansion_method

        if include_imag is None:
            include_imag = [True] * len(carrier_freqs)

        # construct signal approximation function
        def collective_dct(signal_list, t0, n_steps):
            return _signal_list_envelope_DCT(
                signal_list,
                reference_freqs=carrier_freqs,
                degrees=chebyshev_orders,
                t0=t0,
                dt=dt,
                n_intervals=n_steps,
                include_imag=include_imag,
            )

        self._approximate_signals = collective_dct

        # setup rotating frame and single time-step frame change operator
        self._rotating_frame = RotatingFrame(rotating_frame)
        operators = unp.asarray(operators)
        self._operators = operators
        self._Udt = self.rotating_frame.state_out_of_frame(
            dt, np.eye(operators[0].shape[0], dtype=complex)
        )

        perturbations = None

        # set jax-logic dependent components
        if _is_jax_method(integration_method) or _is_diffrax_method(integration_method):
            # compute perturbative terms
            perturbations = _construct_cheb_perturbations_jax(
                operators, chebyshev_orders, carrier_freqs, dt, self.rotating_frame, include_imag
            )
        else:
            perturbations = _construct_cheb_perturbations(
                operators, chebyshev_orders, carrier_freqs, dt, self.rotating_frame, include_imag
            )

        self._dt = dt

        # compute perturbative terms
        results = solve_lmde_perturbation(
            perturbations=perturbations,
            t_span=[0, dt],
            expansion_method=expansion_method,
            expansion_order=expansion_order,
            expansion_labels=expansion_labels,
            integration_method=integration_method,
            **kwargs,
        )

        if self.expansion_method == "dyson":
            # if Dyson multiply by single step frame transformation
            results.perturbation_data.data = self.Udt @ results.perturbation_data.data
            new_coeffs = results.perturbation_data.data[:, -1]
            self._expansion_polynomial = ArrayPolynomial(
                array_coefficients=new_coeffs,
                monomial_labels=results.perturbation_data.labels,
                constant_term=self.Udt,
            )
        elif self.expansion_method == "magnus":
            self._expansion_polynomial = ArrayPolynomial(
                array_coefficients=results.perturbation_data.data[:, -1],
                monomial_labels=results.perturbation_data.labels,
            )

    @property
    def expansion_method(self):
        """Perturbation method used in solver."""
        return self._expansion_method

    @property
    def dt(self):
        """Step size of solver."""
        return self._dt

    @property
    def Udt(self):
        """Single step frame transformation."""
        return self._Udt

    @property
    def operators(self):
        """Original operators in the generator."""
        return self._operators

    @property
    def rotating_frame(self):
        """Rotating frame."""
        return self._rotating_frame

    @property
    def expansion_polynomial(self) -> ArrayPolynomial:
        """ArrayPolynomial object for evaluating the perturbation series."""
        return self._expansion_polynomial

    def approximate_signals(self, signals: List[Signal], t0: float, n_steps: int) -> np.ndarray:
        """Approximate a list of signals over a series of intervals according to the Chebyshev
        time-step and Chebyshev approximation structure set at instantiation.

        Args:
            signals: List of Signals to approximate.
            t0: Start time.
            n_steps: Number of time steps to approximate over.

        Returns:
            np.ndarray: The Chebyshev coefficients of each signal over each time-step.
            The first dimension indexes the signals, and the second the time-step.
        """
        return self._approximate_signals(signals, t0, n_steps)

    def evaluate(self, coeffs: ArrayLike) -> ArrayLike:
        """Evaluate the expansion polynomial for a given list of coefficients."""
        return self.expansion_polynomial(coeffs)


def _construct_cheb_perturbations(
    operators: np.ndarray,
    chebyshev_orders: List[int],
    carrier_freqs: np.ndarray,
    dt: float,
    rotating_frame: RotatingFrame,
    include_imag: Optional[bool] = None,
) -> List[Callable]:
    r"""Helper function for constructing perturbation terms in the expansions used by
    _PerturbativeSolver.

    Constructs a list of operator-valued functions of the form:

    .. math::
        \cos(2 \pi \nu_j t)T_m(t) G_j\textnormal{, and } \sin(-2 \pi \nu_j t)T_m(t) G_j,

    where :math:`\nu_j` and :math:`G_j` are carrier frequency and operator pairs specified in
    ``operators`` and ``carrier_freqs``, and :math:`m \in [0, \dots, r]` where :math:`r` is the
    corresponding integer in ``chebyshev_orders``. The output list is ordered according to the
    lexicographic ordering of the tuples :math:`(j, m)`, with all cosine terms given before sine
    terms.

    Args:
        operators: List of operators.
        chebyshev_orders: List of chebyshev orders for each operator.
        carrier_freqs: List of frequencies for each operator.
        dt: Interval over which the chebyshev polynomials are defined.
        rotating_frame: Frame the operators are in.
        include_imag: List of bools determining whether to keep imaginary components in the signal
            approximation. Defaults to True for all signals. Corresponds to whether or not to
            include the sine terms in this function.

    Returns:
        List of operator-valued functions as described above.
    """

    if include_imag is None:
        include_imag = [True] * len(operators)

    # define functions for constructing perturbations
    def cheb_func(t, deg):
        return _evaluate_cheb_series(t, [0] * deg + [1], domain=[0, dt])

    def get_cheb_func_cos_op(deg, freq, op):
        rad_freq = 2 * np.pi * freq

        def cheb_func_op(t):
            op_in_frame = rotating_frame.operator_into_frame(t, op)
            return cheb_func(t, deg) * np.cos(rad_freq * t) * op_in_frame

        return cheb_func_op

    def get_cheb_func_sin_op(deg, freq, op):
        rad_freq = 2 * np.pi * freq

        def cheb_func_op(t):
            op_in_frame = rotating_frame.operator_into_frame(t, op)
            return cheb_func(t, deg) * np.sin(-rad_freq * t) * op_in_frame

        return cheb_func_op

    # iterate through and construct perturbations list
    perturbations = []
    for deg, op, freq, inc_imag in zip(chebyshev_orders, operators, carrier_freqs, include_imag):
        # construct cosine terms
        for k in range(deg + 1):
            # construct cheb function of appropriate degree
            perturbations.append(get_cheb_func_cos_op(k, freq, op))

        if inc_imag:
            # construct sine terms
            for k in range(deg + 1):
                # construct cheb function of appropriate degree
                perturbations.append(get_cheb_func_sin_op(k, freq, op))

    return perturbations


def _construct_cheb_perturbations_jax(
    operators, chebyshev_orders, carrier_freqs, dt, rotating_frame, include_imag=None
):
    """JAX version of _construct_cheb_perturbations."""

    if include_imag is None:
        include_imag = [True] * len(operators)

    # define functions for constructing perturbations list
    def get_cheb_func(deg):
        c = jnp.array([0] * deg + [1], dtype=float)

        def cheb_func(t):
            return _evaluate_cheb_series_jax(t, c, domain=[0, dt])

        return cheb_func

    def get_cheb_func_cos_op(deg, freq, op):
        rad_freq = 2 * np.pi * freq
        cheb_func = get_cheb_func(deg)

        def cheb_func_op(t):
            op_in_frame = rotating_frame.operator_into_frame(t, op)
            return cheb_func(t) * jnp.cos(rad_freq * t) * op_in_frame

        return cheb_func_op

    def get_cheb_func_sin_op(deg, freq, op):
        rad_freq = 2 * np.pi * freq
        cheb_func = get_cheb_func(deg)

        def cheb_func_op(t):
            op_in_frame = rotating_frame.operator_into_frame(t, op)
            return cheb_func(t) * jnp.sin(-rad_freq * t) * op_in_frame

        return cheb_func_op

    # iterate through and construct perturbations list
    perturbations = []
    for deg, op, freq, inc_imag in zip(chebyshev_orders, operators, carrier_freqs, include_imag):
        # construct cosine terms
        for k in range(deg + 1):
            # construct cheb function of appropriate degree
            perturbations.append(get_cheb_func_cos_op(k, freq, op))

        if inc_imag:
            # construct sine terms
            for k in range(deg + 1):
                # construct cheb function of appropriate degree
                perturbations.append(get_cheb_func_sin_op(k, freq, op))

    return perturbations


def _evaluate_cheb_series(
    x: Union[float, np.ndarray], c: np.ndarray, domain: Optional[List] = None
) -> Union[float, np.ndarray]:
    """Evaluate a Chebyshev series on a given domain.

    This calls ``numpy.polynomial.chebyshev.chebval`` but on a stretched domain.

    Args:
        x: Array of x values to evaluate the Chebyshev series on.
        c: Array of Chebyshev coefficients.
        domain: Domain over which the the Chebyshev series is defined. Defaults to [-1, 1].

    Returns:
        array: Chebyshev series evaluated on x.
    """
    domain = domain or [-1, 1]
    x = (2 * x - domain[1] - domain[0]) / (domain[1] - domain[0])
    return chebval(x, c)


def _evaluate_cheb_series_jax(
    x: Union[float, np.ndarray], c: np.ndarray, domain: Optional[List] = None
) -> Union[float, np.ndarray]:
    """Evaluate Chebyshev series on a on a given domain using JAX looping logic.
    This follows the same algorithm as ``numpy.polynomial.chebyshev.chebval`` but uses JAX looping
    logic.

    Args:
        x: Array of x values to evaluate the Chebyshev series on.
        c: Array of Chebyshev coefficients.
        domain: Domain over which the the Chebyshev series is defined. Defaults to [-1, 1].

    Returns:
        array: Chebyshev series evaluated on x.
    """
    domain = domain or [-1, 1]

    x = (2 * x - domain[1] - domain[0]) / (domain[1] - domain[0])

    if len(c) == 1:
        c0 = c[0]
        c1 = 0
    elif len(c) == 2:
        c0 = c[0]
        c1 = c[1]
    else:
        x2 = 2 * x

        def scan_fun(carry, c_val):
            c0, c1 = carry
            tmp = c0
            c0 = c_val - c1
            c1 = tmp + c1 * x2
            return (c0, c1), None

        init_c1 = c[-1] * jnp.ones_like(x)
        init_c2 = c[-2] * jnp.ones_like(x)
        c0, c1 = scan(scan_fun, init=(init_c2, init_c1), xs=jnp.flip(c)[2:])[0]

    return c0 + c1 * x


def _signal_list_envelope_DCT(
    signal_list: List[Signal],
    reference_freqs: ArrayLike,
    degrees: List[int],
    t0: float,
    dt: float,
    n_intervals: int,
    include_imag: Optional[List] = None,
) -> ArrayLike:
    """Compute envelope DCT on a list of signals, and compile results into a single list,
    separating real and imaginary.

    Args:
        signal_list: List of signals whose envelopes the DCT will be performed on.
        reference_freqs: Reference frequencies for defining the envelopes.
        degrees: List of Chebyshev degrees for doing the approximation (one for each signal).
        t0: Start time.
        dt: Time step size.
        n_intervals: Number of Intervals.
        include_imag: Whether or not to include the imaginary components of the envelopes. Defaults
            to ``True`` for all.

    Returns:
        Discrete Chebyshev transform of the envelopes with reference frequencies.
    """

    if include_imag is None:
        include_imag = [True] * len(signal_list)

    def envelope_DCT(sig, freq, degree):
        return _signal_envelope_DCT(sig, freq, degree, t0, dt, n_intervals)

    # initialize coefficient array with first signal
    coeffs = envelope_DCT(signal_list[0], reference_freqs[0], degrees[0])
    if include_imag[0]:
        coeffs = unp.append(coeffs.real, coeffs.imag, axis=0)
    else:
        coeffs = coeffs.real

    for sig, freq, deg, inc_imag in zip(
        signal_list[1:], reference_freqs[1:], degrees[1:], include_imag[1:]
    ):
        new_coeffs = envelope_DCT(sig, freq, deg)

        coeffs = unp.append(coeffs, new_coeffs.real, axis=0)
        if inc_imag:
            coeffs = unp.append(coeffs, new_coeffs.imag, axis=0)

    return coeffs


def _signal_envelope_DCT(
    signal: Signal, reference_freq: float, degree: int, t0: float, dt: float, n_intervals: int
) -> ArrayLike:
    """Perform multi-interval DCT on the envelope of a Signal relative to a reference frequency.
    I.e. This is equivalent to shifting the frequency of the signal to the reference frequency,
    and performing the multi-interval DCT on the resultant envelope.

    Args:
        signal: Signal to approximate.
        reference_freq: Reference frequency to shift the signal frequency to.
        degree: Degree of Chebyshev approximation.
        t0: Start time.
        dt: Length of intervals.
        n_intervals: Number of intervals to perform DCT over.

    Returns:
        ArrayLike: Multi-interval DCT stored as a 2d array.
    """

    t_vals = t0 + np.arange(n_intervals) * dt
    phase_arg = -1j * 2 * np.pi * reference_freq
    final_phase_shift = unp.exp(-phase_arg * t_vals)

    def shifted_env(t):
        return signal.complex_value(t) * np.exp(phase_arg * t)

    return _multi_interval_DCT(shifted_env, degree, t0, dt, n_intervals) * unp.expand_dims(
        final_phase_shift, axis=0
    )


def _multi_interval_DCT(
    f: Callable, degree: int, t0: float, dt: float, n_intervals: int
) -> ArrayLike:
    """Evaluate the multi-interval DCT of a function f over contiguous intervals of size dt
    starting at t0.

    Note: This function assumes that ``f`` is vectorized, and the returned array is indexed as:
        - First axis is coefficient.
        - Second axis is interval.

    Args:
        f: Vectorized function to approximate.
        degree: Degree of Chebyshev approximation.
        t0: Start time.
        dt: Length of intervals.
        n_intervals: Number of intervals to perform DCT over.

    Returns:
        ArrayLike: Multi-interval DCT stored as a 2d array.
    """

    dct_mat, xcheb = _construct_DCT(degree, domain=[0, dt])

    # compute all times at which the function needs to be evaluated
    interval_start_times = t0 + np.arange(n_intervals) * dt

    # time values: columns correspond to interval, rows are the shifted chebyshev values
    x_vals = np.add.outer(xcheb, interval_start_times)

    return dct_mat @ f(x_vals)


def _construct_DCT(degree: int, domain: Optional[List] = None) -> Tuple:
    """Construct the matrix and evaluation points for performing the Discrete Chebyshev
    Transform (DCT) over an interval specified by ``domain``. This utilizes code from
    :mod:`numpy.polynomial.chebyshev.chebinterpolate`, but modified to allow for interval
    specification, and to return the constructed matrices and evaluation points for performing the
    DCT.

    I.e. For outputs ``(M, x)``, the DCT of a function ``f`` over the specified domain can be
    evaluated as ``M @ f(x)``.

    Args:
        degree: Degree of Chebyshev approximation.
        domain: Interval of approximation. Defaults to [-1, 1].

    Returns:
        Tuple: Pair of arrays for performing the DCT.
    """
    domain = domain or [-1, 1]
    order = degree + 1

    xcheb = chebpts1(order)
    xcheb_shifted = 0.5 * ((domain[1] - domain[0]) * xcheb + (domain[1] + domain[0]))

    dct_mat = chebvander(xcheb, degree).T
    dct_mat[0] /= order
    dct_mat[1:] /= 0.5 * order

    return dct_mat, xcheb_shifted



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Register custom functions using alias
"""

from .asarray import register_asarray
from .matmul import register_matmul
from .rmatmul import register_rmatmul
from .multiply import register_multiply
from .linear_combo import register_linear_combo
from .conjugate import register_conjugate
from .transpose import register_transpose



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
Class for storing results of perturbation theory computations.
"""

from typing import List, Mapping, Optional
from dataclasses import dataclass

from multiset import Multiset

from qiskit import QiskitError


@dataclass
class _LabeledData:
    """Container for arbitrarily "labeled data", i.e. data whose indices are arbitrary python
    objects. The method ``get_item`` looks up an item according to the label.
    """

    data: Mapping[int, any]
    labels: List[any]
    metadata: Optional[any] = None

    def get_item(self, label: any) -> any:
        """Look up an item in self.data according to the location of label in self.labels."""
        label = self._preprocess_label(label)

        if label in self.labels:
            return self.data[self.labels.index(label)]

        raise QiskitError("label is not present in self.labels.")

    def _preprocess_label(self, label: any) -> any:
        return label


class PowerSeriesData(_LabeledData):
    """Storage container for power series data. Labels are assumed to be ``Multiset`` instances, and
    data is assumed to be a dense ``ArrayLike``.
    """

    def _preprocess_label(self, label: Multiset) -> Multiset:
        """Cast to a Multiset."""
        return Multiset(label)


class DysonLikeData(_LabeledData):
    """Storage container for DysonLike series data. Labels are assumed to be lists of ints, and data
    is assumed to be a dense ``ArrayLike``.
    """

    def _preprocess_label(self, label: list) -> list:
        """Cast to a list."""
        return list(label)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
========================================
Systems (:mod:`qiskit_dynamics.systems`)
========================================

.. currentmodule:: qiskit_dynamics.systems

This module provides high level interfaces for building and solving models of quantum systems. Where
the :mod:`.solvers` and :mod:`.models` modules provide interfaces for defining and solving systems
in terms of user-defined arrays, this module provides tools for building descriptions of systems in
terms of tensor-factor subsystems, an algebraic system for defining operators on subsystems, a high
level class representing an abstract dynamical model of a quantum system, and tools for analysing
results. The ultimate purpose of the module is to minimize the need for a user to work explicitly
with building and manipulating arrays and array indexing, which can be time consuming and prone to
error. See the :ref:`Systems Modelling Tutorial <systems modelling tutorial>` and the :ref:`How-to
use advanced system modelling functionality <systems modelling userguide>` for detailed examples.

The core building block of a model is a :class:`.Subsystem`, which represents a single
finite-dimensional complex vector space on which to define the model of a quantum system. A single
model may be defined on multiple subsystems, in which each subsystem represents a tensor factor in a
tensor-product space.

.. code-block:: python

    Q0 = Subsystem(name="Q0", dim=2)
    Q1 = Subsystem(name="Q0", dim=2)


Abstract operators acting on these subsystems can be defined as follows:

.. code-block:: python

    X0 = X(Q0)
    Y1 = Y(Q1)

Using algebraic operations, new operators may be defined. For example, the tensor product of ``X``
on ``Q0`` and ``Y`` on ``Q1`` can be constructed through matrix multiplication:

.. code-block:: python

    X0 @ Y1

Similarly, the sum of these operators can be constructed through addition ``X0 + Y1``. To facilitate
working with operators on subsystems without needing to always specify the full context of all
subsystems in a given model, operators are always assumed to act as the identity on all unspecified
subsystems, similar to the common mathematical notation in which an operator :math:`a_2` means "the
operator :math:`a` acting on subsystem :math:`2` and the identity on all others".

The matrix of an abstract operator can be built by calling the ``matrix`` method. The specific
ordering of the tensor factors desired can be supplied, e.g.:

.. code-block:: python

    (X0 @ Y1).matrix(ordered_subsystems=[Q0, Q1])

If no explicitly ordering is supplied, the default internal ordering built during the construction
of the operator will be used. In addition to a set of pre-defined operators, users can instantiate a
:class:`.SubsystemOperator` instance with an arbitrary concrete matrix which acts on an arbitrary
list of :class:`.Subsystem` instances.

Operators can be assumbled into models of quantum systems using the :class:`.QuantumSystemModel`
class. For example, a model of a standard qubit can be built as follows:

.. code-block:: python

    q0_model = QuantumSystemModel(
        static_hamiltonian=2 * np.pi * 5. * N(Q0),
        drive_hamiltonians=[2 * np.pi * 0.1 * X(Q0)],
        drive_hamiltonian_coefficients=["d0"]
    )

This model can now be solved with a single call:

.. code-block:: python

    results = q0_model.solve(
        signals={"d0": Signal(1., carrier_freq=5.)},
        t_span=t_span,
        t_eval=t_eval,
        y0=y0
    )

with ``results`` being the standard ``OdeResult`` object returned by Qiskit Dynamics solvers.

In addition to the functionality above, this module contains the :class:`SubsystemMapping` class for
defining linear maps between tensor factor spaces given as lists of :class:`Subsystem` instances. As
shown in the :ref:`How-to use advanced system modelling functionality <systems modelling userguide>`
userguide entry, this class can be used to define injections of subspaces into larger spaces, or to
restrict a model to a subspace of interest.

Furthermore, the :class:`ONBasis` and :class:`DressedBasis` classes represent bases for subspaces on
tensor product spaces represented by lists of :class:`Subsystem` instances.

System modelling classes
========================

.. autosummary::
   :toctree: ../stubs/

   Subsystem
   SubsystemOperator
   FunctionOperator
   ONBasis
   DressedBasis
   SubsystemMapping
   QuantumSystemModel
   IdealQubit
   DuffingOscillator
   ExchangeInteraction

Pre-defined operators
=====================

.. autosummary::
   :toctree: ../stubs/

   I
   X
   Y
   Z
   N
   A
   Adag
"""

from .subsystem import Subsystem
from .subsystem_operators import *
from .abstract_subsystem_operators import FunctionOperator
from .orthonormal_basis import ONBasis, DressedBasis
from .subsystem_mapping import SubsystemMapping
from .quantum_system_model import *



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Module for transformations between signals.
"""

from abc import ABC, abstractmethod
from typing import Callable, Union, List
import numpy as np

from qiskit import QiskitError
from qiskit_dynamics.arraylias import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias.alias import _numpy_multi_dispatch

from .signals import Signal, DiscreteSignal


class BaseTransferFunction(ABC):
    """Base class for transforming signals."""

    @property
    @abstractmethod
    def n_inputs(self):
        """Number of input signals to the transfer function."""
        pass

    def __call__(self, *args, **kwargs) -> Union[Signal, List[Signal]]:
        """Apply the transfer function to the input signals.

        Args:
            *args: The signals to which the transfer function will be applied.
            **kwargs: Key word arguments to control the transfer functions.

        Returns:
            Signal: The transformed signal.

        Raises:
            QiskitError: if the number of args is not correct.
        """

        if len(args) != self.n_inputs:
            raise QiskitError(
                f"""{self.__class__.__name__} expected {len(args)}
                input signals but {self.n_inputs} were given."""
            )

        return self._apply(*args, **kwargs)

    @abstractmethod
    def _apply(self, *args, **kwargs) -> Union[Signal, List[Signal]]:
        """Applies a transformation on a signal, such as a convolution, low pass filter, etc.

        Args:
            *args: The signals to which the transfer function will be applied.
            **kwargs: Key word arguments to control the transfer functions.

        Returns:
            Signal: The transformed signal.
        """
        pass


class Convolution(BaseTransferFunction):
    """Applies a convolution as a sum

        (f*g)(n) = sum_k f(k)g(n-k)

    The implementation is quadratic in the number of samples in the signal.
    """

    def __init__(self, func: Callable):
        """
        Args:
            func: The convolution function specified in time. This function will be normalized to
                one before doing the convolution. To scale signals multiply them by a float.
        """
        self._func = func

    @property
    def n_inputs(self):
        return 1

    # pylint: disable=arguments-differ
    def _apply(self, signal: Signal) -> Signal:
        """Applies a transformation on a signal, such as a convolution, low pass filter, etc. Once a
        convolution is applied the signal can longer have a carrier as the carrier is part of the
        signal value and gets convolved.

        Args:
            signal: A signal or list of signals to which the transfer function will be applied.

        Returns:
            signal: The transformed signal or list of signals.

        Raises:
            QiskitError: if the signal is not pwc.
        """
        if isinstance(signal, DiscreteSignal):
            # Perform a discrete time convolution.
            dt = signal.dt
            func_samples = unp.asarray([self._func(dt * i) for i in range(signal.duration)])
            func_samples = func_samples / unp.sum(func_samples)
            sig_samples = signal(dt * unp.arange(signal.duration))

            convoluted_samples = _numpy_multi_dispatch(func_samples, sig_samples, path="convolve")

            return DiscreteSignal(dt, convoluted_samples, carrier_freq=0.0, phase=0.0)
        else:
            raise QiskitError("Transfer function not defined on input.")


class FFTConvolution(BaseTransferFunction):
    """Applies a convolution by moving into the fourier domain."""

    def __init__(self, func: Callable):
        self._func = func

    @property
    def n_inputs(self):
        return 1

    # pylint: disable=arguments-differ
    def _apply(self, signal: Signal) -> Signal:
        raise NotImplementedError


class Sampler(BaseTransferFunction):
    """Resample a signal by wrapping DiscreteSignal.from_Signal."""

    def __init__(self, dt: float, n_samples: int, start_time: float = 0):
        """
        Args:
            dt: The new sample period.
            n_samples: number of samples to resample with.
            start_time: start time from which to resample.
        """
        self._dt = dt
        self._n_samples = n_samples
        self._start_time = start_time

    @property
    def n_inputs(self):
        """Number of input signals to the transfer function."""
        return 1

    # pylint: disable=arguments-differ
    def _apply(self, signal: Signal) -> Signal:
        """Apply the transfer function to the signal."""
        return DiscreteSignal.from_Signal(
            signal, dt=self._dt, n_samples=self._n_samples, start_time=self._start_time
        )


class IQMixer(BaseTransferFunction):
    """Implements an IQ Mixer.

    The IQ mixer takes as input three signals:
    - in-phase signal: I cos(w_if t + phi_I)
    - quadrature: Q cos(w_if t + phi_Q)
    - local oscillator: K cos(w_lo t)

    In this implementation the local oscillator is specified by its frequency w_lo and, without loss
    of generality we assume K = 1. Furthermore, we require that the carrier frequency of the I and Q
    be identical.

    The output RF signal is defined by

    s_rf = I [cos(wp t + phi_I) + cos(wm t + phi_I)]/2
         + Q [cos(wp t + phi_Q - pi/2) + cos(wm t + phi_Q + pi/2)]/2

    where wp = w_lo + w_if and wp = w_lo - w_if.

    The output of this transfer function will produce a piece-wise constant that does not have a
    carrier frequency or phase. All information is in the samples. Mixer imperfections are not
    included.
    """

    def __init__(self, lo: float):
        """
        Args:
            lo: The carrier of the IQ mixer.
        """
        self._lo = lo

    @property
    def n_inputs(self):
        return 2

    # pylint: disable=arguments-differ
    def _apply(self, si: Signal, sq: Signal) -> Signal:
        """
        Args:
            si: In phase signal
            sq: Quadrature signal.

        Returns:
            The up-converted signal.

        Raises:
            QiskitError: if the carriers frequencies of I and Q differ.
        """

        # Check compatibility of the input signals
        if si.carrier_freq != sq.carrier_freq:
            raise QiskitError("IQ mixer requires the same sideband frequencies for I and Q.")

        phi_i, phi_q = si.phase, sq.phase
        wp, wm = self._lo + si.carrier_freq, self._lo - si.carrier_freq
        wp *= 2 * np.pi
        wm *= 2 * np.pi

        def mixer_func(t):
            """Function of the IQ mixer."""
            osc_i = unp.cos(wp * t + phi_i) + unp.cos(wm * t + phi_i)
            osc_q = unp.cos(wp * t + phi_q - np.pi / 2) + unp.cos(wm * t + phi_q + np.pi / 2)
            return si.envelope(t) * osc_i / 2 + sq.envelope(t) * osc_q / 2

        return Signal(mixer_func, carrier_freq=0, phase=0)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import setuptools

requirements = [
    "numpy>=1.17",
    "scipy>=1.4",
    "matplotlib>=3.0",
    "qiskit<=1.3",
    "multiset>=3.0.1",
    "sympy>=1.12",
    "arraylias"
]

jax_extras = ['jax', 'jaxlib']

PACKAGES = setuptools.find_packages(exclude=['test*'])

version_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'qiskit_dynamics',
                 'VERSION.txt'))

with open(version_path, 'r') as fd:
    version = fd.read().rstrip()

README_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                           'README.md')
with open(README_PATH) as readme_file:
    README = readme_file.read()

setuptools.setup(
    name="qiskit-dynamics",
    version=version,
    packages=PACKAGES,
    description="Qiskit ODE solver",
    long_description=README,
    long_description_content_type='text/markdown',
    url="https://github.com/Qiskit-Community/qiskit-dynamics",
    author="Qiskit Development Team",
    author_email="qiskit@us.ibm.com",
    license="Apache 2.0",
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering",
    ],
    keywords="qiskit sdk quantum",
    project_urls={
        "Bug Tracker": "https://github.com/Qiskit-Community/qiskit-dynamics/issues",
        "Source Code": "https://github.com/Qiskit-Community/qiskit-dynamics",
        "Documentation": "https://qiskit-community.github.io/qiskit-dynamics/",
    },
    install_requires=requirements,
    include_package_data=True,
    python_requires=">=3.10",
    extras_require={
        "jax": jax_extras
    },
    zip_safe=False
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Array polynomial.
"""

from typing import List, Optional, Callable, Tuple, Union
from copy import copy
from itertools import product

from multiset import Multiset

import numpy as np
from numpy.typing import DTypeLike

from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import DYNAMICS_NUMPY_ALIAS as numpy_alias
from qiskit_dynamics.arraylias.alias import (
    _preferred_lib,
    _numpy_multi_dispatch,
    ArrayLike,
    _isArrayLike,
)

from qiskit_dynamics.perturbation.multiset_utils import (
    _validate_non_negative_ints,
    _get_all_submultisets,
    _sorted_multisets,
    _submultisets_and_complements,
    _multiset_to_sorted_list,
)
from qiskit_dynamics.perturbation.custom_binary_op import _CustomBinaryOp

try:
    import jax.numpy as jnp
except ImportError:
    pass


class ArrayPolynomial:
    r"""A polynomial with array-valued coefficients.

    This class represents a multi-variable function of the form:

    .. math::
        f(c_1, \dots, c_r) = A_\emptyset + \sum_{I \in S} c_I A_I,

    where in the above:

        - :math:`S` is a finite set of multisets indicating non-zero monomial terms,
        - For a given multiset of non-negative integers :math:`I=(i_1, \dots, i_k)`,
          :math:`c_I = c_{i_1} \times \dots \times c_{i_k}`, and
        - The :math:`A_I` are arrays of the same shape, indexed by the first dimension.

    See the :ref:`multiset and power series notation section <multiset power series>` of the
    perturbation review for an explanation of the multiset notation.

    An :class:`.ArrayPolynomial` is instantiated with the arguments:

        - ``constant_term`` specifying the array :math:`A_\emptyset`.
        - ``array_coefficients`` specifying a list of the arrays :math:`A_I`, or as a single array
          whose first index lists the :math:`A_I`,
        - ``monomial_labels`` specifying the set :math:`S` as a list of ``Multiset`` instances
          ordered in correspondence with ``array_coefficients``.

    For example, the :class:`.ArrayPolynomial` corresponding to the mathematical polynomial

    .. math::

        f(c_0, c_1) = A_\emptyset
            + c_{(0)} A_{(0)} + c_{(0, 1)}A_{(0, 1)} + c_{(1, 1)}A_{(1, 1)}

    for arrays :math:`A_\emptyset, A_{(0)}, A_{(0, 1)}, A_{(1, 1)}` stored in variables ``A_c``,
    ``A0``, ``A01``, and ``A11`` can be instantiated with

    .. code-block:: python

        ap = ArrayPolynomial(
            constant_term = A_c
            array_coefficients=[A0, A01, A11],
            monomial_labels=[Multiset({0: 1}), Multiset({0: 1, 1: 1}), Multiset({1: 2})]
        )

    Once instantiated, the polynomial can be evaluated on an array of variable values, e.g.

    .. code-block:: python

        c = np.array([c0, c1])
        ap(c) # polynomial evaluated on variables

    :class:`.ArrayPolynomial` supports some array properties, e.g. ``ap.shape`` and ``ap.ndim``
    return the shape and number of dimensions of the output of the polynomial. Some array methods
    are also supported, such as ``transpose`` and ``trace``, and their output produces a new
    :class:`.ArrayPolynomial` which evaluates to the array one would obtain by first evaluating the
    original, then calling the array method. E.g.

    .. code-block:: python

        ap2 = ap1.transpose()
        ap2(c) == ap1(c).transpose()

    Finally, :class:`.ArrayPolynomial` supports algebraic operations, e.g.

    .. code-block:: python

        ap3 = ap1 @ ap2
        ap3(c) == ap1(c) @ ap2(c)

    It also has specialized algebraic methods that perform algebraic operations while "ignoring"
    terms. E.g., for two instances ``ap1`` and ``ap2``, the call

    .. code-block:: python

        ap1.matmul(ap2, monomial_filter=lambda x: len(x) <= 3)

    is similar to ``ap1 @ ap2``, but will result in an :class:`.ArrayPolynomial` in which all terms
    of degree larger than ``3`` will not be included in the results.
    """

    __array_priority__ = 20

    def __init__(
        self,
        constant_term: Optional[ArrayLike] = None,
        array_coefficients: Optional[ArrayLike] = None,
        monomial_labels: Optional[List[Multiset]] = None,
        array_library: Optional[str] = None,
    ):
        """Construct a multivariable matrix polynomial.

        Args:
            constant_term: An array representing the constant term of the polynomial.
            array_coefficients: A 3d array representing a list of array coefficients.
            monomial_labels: A list of multisets with non-negative integer entries of the same
                length as ``array_coefficients`` indicating the monomial coefficient for each
                corresponding ``array_coefficients``.
            array_library: Array library for stored array terms.
        Raises:
            QiskitError: If insufficient information is supplied to define an ArrayPolynomial,
                or if monomial labels contain anything other than non-negative integers.
        """

        if array_coefficients is None and constant_term is None:
            raise QiskitError(
                "At least one of array_coefficients and constant_term must be specified."
            )

        if monomial_labels is not None:
            self._monomial_labels = [Multiset(m) for m in monomial_labels]
            for m in self._monomial_labels:
                _validate_non_negative_ints(m)
        else:
            self._monomial_labels = []

        self._array_coefficients = None
        if array_coefficients is not None:
            self._array_coefficients = numpy_alias(like=array_library).asarray(array_coefficients)

        self._constant_term = None
        if constant_term is not None:
            self._constant_term = numpy_alias(like=array_library).asarray(constant_term)

        # there is redundancy here as both perform pre-computation on self._monomial_labels
        self._compute_monomials = _get_monomial_compute_function(self._monomial_labels)
        self._compute_monomials_jax = _get_monomial_compute_function_jax(self._monomial_labels)

    @property
    def monomial_labels(self) -> Union[List, None]:
        """The monomial labels corresponding to non-constant terms."""
        return self._monomial_labels

    @property
    def array_coefficients(self) -> Union[ArrayLike, None]:
        """The array coefficients for non-constant terms."""
        return self._array_coefficients

    @property
    def constant_term(self) -> Union[ArrayLike, None]:
        """The constant term."""
        return self._constant_term

    def compute_monomials(self, c: ArrayLike) -> ArrayLike:
        """Vectorized computation of all scalar monomial terms in the polynomial as specified by
        ``self.monomial_labels``.

        Args:
            c: Array of variables.
        Returns:
            Array of all monomial terms ordered according to ``self.monomial_labels``.
        """
        lib = _preferred_lib(c, self.constant_term, self.array_coefficients)

        if "jax" in lib:
            return self._compute_monomials_jax(c)

        return self._compute_monomials(c)

    @property
    def shape(self) -> Tuple[int]:
        """Shape of the arrays in the polynomial."""
        if self._constant_term is not None:
            return self._constant_term.shape
        else:
            return self._array_coefficients.shape[1:]

    @property
    def ndim(self) -> int:
        """Number of dimensions of the coefficients of the polynomial."""
        if self._constant_term is not None:
            return self._constant_term.ndim
        else:
            return self._array_coefficients.ndim - 1

    def conj(self) -> "ArrayPolynomial":
        """Return an ArrayPolynomial that is the conjugate of self."""

        constant_term = None
        coefficients = None

        if self._constant_term is not None:
            constant_term = unp.conj(self._constant_term)

        if self._array_coefficients is not None:
            coefficients = unp.conj(self._array_coefficients)

        return ArrayPolynomial(
            array_coefficients=coefficients,
            monomial_labels=copy(self._monomial_labels),
            constant_term=constant_term,
        )

    def transpose(self, axes: Optional[Tuple[int]] = None) -> "ArrayPolynomial":
        """Return the ArrayPolynomial when transposing all coefficients."""

        constant_term = None
        coefficients = None

        if self._constant_term is not None:
            constant_term = unp.transpose(self._constant_term, axes)

        if self._array_coefficients is not None:
            if axes is None:
                axes = tuple(range(1, self.ndim + 1)[::-1])
            else:
                axes = tuple(ax + 1 for ax in axes)
            axes = (0,) + axes
            coefficients = unp.transpose(self._array_coefficients, axes)

        return ArrayPolynomial(
            array_coefficients=coefficients,
            monomial_labels=copy(self._monomial_labels),
            constant_term=constant_term,
        )

    def trace(
        self,
        offset: Optional[int] = 0,
        axis1: Optional[int] = 0,
        axis2: Optional[int] = 1,
        dtype: Optional[DTypeLike] = None,
    ) -> "ArrayPolynomial":
        """Take the trace of the coefficients."""

        if self.ndim < 2:
            raise QiskitError("ArrayPolynomial.trace() requires ArrayPolynomial.ndim at least 2.")

        constant_term = None
        coefficients = None

        if self._constant_term is not None:
            constant_term = unp.trace(
                self._constant_term, offset=offset, axis1=axis1, axis2=axis2, dtype=dtype
            )

        if self._array_coefficients is not None:
            coefficients = unp.trace(
                self._array_coefficients,
                offset=offset,
                axis1=axis1 + 1,
                axis2=axis2 + 1,
                dtype=dtype,
            )

        return ArrayPolynomial(
            array_coefficients=coefficients,
            monomial_labels=copy(self._monomial_labels),
            constant_term=constant_term,
        )

    def sum(self, axis: Optional[Union[int, Tuple[int]]] = None, dtype: Optional[DTypeLike] = None):
        """Perform a sum on the coefficients."""

        constant_term = None
        coefficients = None

        # constant term can be handled normally
        if self.constant_term is not None:
            constant_term = self.constant_term.sum(axis=axis, dtype=dtype)

        # axis must be shifted for array coefficients
        if self.array_coefficients is not None:
            if self.ndim == 0 and axis is None:
                coefficients = unp.array(self.array_coefficients, dtype=dtype)
            else:
                if axis is None:
                    axis = tuple(k for k in range(1, self.ndim + 1))
                elif isinstance(axis, int):
                    axis = axis + 1
                elif isinstance(axis, tuple):
                    axis = tuple(k + 1 for k in axis)

                coefficients = self.array_coefficients.sum(axis=axis, dtype=dtype)

        return ArrayPolynomial(
            array_coefficients=coefficients,
            monomial_labels=copy(self._monomial_labels),
            constant_term=constant_term,
        )

    @property
    def real(self) -> "ArrayPolynomial":
        """Return the real part of self."""

        constant_term = None
        array_coefficients = None

        if self.constant_term is not None:
            constant_term = self.constant_term.real

        if self.array_coefficients is not None:
            array_coefficients = self.array_coefficients.real

        return ArrayPolynomial(
            array_coefficients=array_coefficients,
            monomial_labels=copy(self._monomial_labels),
            constant_term=constant_term,
        )

    def add(
        self,
        other: Union["ArrayPolynomial", ArrayLike],
        monomial_filter: Optional[Callable] = None,
    ) -> "ArrayPolynomial":
        """Add two polynomials with bounds on which terms to keep.

        Optionally, a function ``monomial_filter`` can be provided to limit which monomials appear
        in the output. It must accept as input a ``Multiset`` and return a ``bool``, and a term with
        label given by ``multiset`` will be included only if ``monomial_filter(multiset) == True``,
        and will not be computed if ``monomial_filter(multiset) == False``.

        Args:
            other: Other to add to self.
            monomial_filter: Function determining which terms to compute and keep.
        Returns:
            ArrayPolynomial achieved by adding both self and other.
        Raises:
            QiskitError: if other cannot be cast as an ArrayPolynomial.
        """

        if _isArrayLike(other):
            other = ArrayPolynomial(constant_term=other)

        if isinstance(other, ArrayPolynomial):
            return _array_polynomial_addition(self, other, monomial_filter=monomial_filter)

        raise QiskitError(
            "Only types castable as an ArrayPolynomial can be added to an ArrayPolynomial."
        )

    def matmul(
        self,
        other: Union["ArrayPolynomial", ArrayLike],
        monomial_filter: Optional[Callable] = None,
    ) -> "ArrayPolynomial":
        """Matmul self @ other with bounds on which terms to keep.

        Optionally, a function ``monomial_filter`` can be provided to limit which monomials appear
        in the output. It must accept as input a ``Multiset`` and return a ``bool``, and a term with
        label given by ``multiset`` will be included only if ``monomial_filter(multiset) == True``,
        and will not be computed if ``monomial_filter(multiset) == False``.

        Args:
            other: Other to add to self.
            monomial_filter: Function determining which terms to compute and keep.
        Returns:
            ArrayPolynomial achieved by matmul of self and other.
        Raises:
            QiskitError: if other cannot be cast as an ArrayPolynomial.
        """
        if _isArrayLike(other):
            other = ArrayPolynomial(constant_term=other)

        if isinstance(other, ArrayPolynomial):
            return _array_polynomial_distributive_binary_op(
                self, other, lambda A, B: A @ B, monomial_filter=monomial_filter
            )

        raise QiskitError(f"Type {type(other)} not supported by ArrayPolynomial.matmul.")

    def mul(
        self,
        other: Union["ArrayPolynomial", ArrayLike],
        monomial_filter: Optional[Callable] = None,
    ) -> "ArrayPolynomial":
        """Entrywise multiplication of two ArrayPolynomials with bounds on which terms to keep.

        Optionally, a function ``monomial_filter`` can be provided to limit which monomials appear
        in the output. It must accept as input a ``Multiset`` and return a ``bool``, and a term with
        label given by ``multiset`` will be included only if ``monomial_filter(multiset) == True``,
        and will not be computed if ``monomial_filter(multiset) == False``.

        Args:
            other: Other to add to self.
            monomial_filter: Function determining which terms to compute and keep.
        Returns:
            ArrayPolynomial achieved by matmul of self and other.
        Raises:
            QiskitError: if other cannot be cast as an ArrayPolynomial.
        """

        if _isArrayLike(other):
            other = ArrayPolynomial(constant_term=other)

        if isinstance(other, ArrayPolynomial):
            return _array_polynomial_distributive_binary_op(
                self, other, lambda A, B: A * B, monomial_filter=monomial_filter
            )

        raise QiskitError(f"Type {type(other)} not supported by ArrayPolynomial.mul.")

    def __add__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        """Dunder method for addition of two ArrayPolynomials."""
        return self.add(other)

    def __radd__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        """Dunder method for right-addition of two ArrayPolynomials."""
        return self.add(other)

    def __neg__(self) -> "ArrayPolynomial":
        constant_term = None
        if self.constant_term is not None:
            # pylint: disable=invalid-unary-operand-type
            constant_term = -self.constant_term

        array_coefficients = None
        if self.array_coefficients is not None:
            # pylint: disable=invalid-unary-operand-type
            array_coefficients = -self.array_coefficients

        return ArrayPolynomial(
            constant_term=constant_term,
            monomial_labels=self.monomial_labels,
            array_coefficients=array_coefficients,
        )

    def __sub__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        return self + (-other)

    def __rsub__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        return other + (-self)

    def __mul__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        """Dunder method for entry-wise multiplication."""
        return self.mul(other)

    def __rmul__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        """Dunder method for right-multiplication."""
        return self.mul(other)

    def __matmul__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        """Dunder method for matmul."""
        return self.matmul(other)

    def __rmatmul__(self, other: Union["ArrayPolynomial", ArrayLike]) -> "ArrayPolynomial":
        """Dunder method for rmatmul."""
        if _isArrayLike(other):
            other = ArrayPolynomial(constant_term=other)

        if isinstance(other, ArrayPolynomial):
            return other.matmul(self)

        raise QiskitError(f"Type {type(other)} not supported by ArrayPolynomial.__rmatmul__.")

    def __getitem__(self, idx) -> "ArrayPolynomial":
        """Index the ArrayPolynomial similarly to an array."""
        constant_term = None
        array_coefficients = None

        if self.constant_term is not None:
            constant_term = self.constant_term[idx]

        if self.array_coefficients is not None:
            array_coefficients = self.array_coefficients[(slice(None),) + idx]

        return ArrayPolynomial(
            array_coefficients=array_coefficients,
            monomial_labels=copy(self._monomial_labels),
            constant_term=constant_term,
        )

    def __len__(self) -> int:
        """Number of terms in the polynomial."""
        num_terms = 0
        if self.array_coefficients is not None:
            num_terms = num_terms + len(self.array_coefficients)

        if self.constant_term is not None:
            num_terms = num_terms + 1

        return num_terms

    def __call__(self, c: Optional[ArrayLike] = None) -> ArrayLike:
        """Evaluate the polynomial.

        Args:
            c: Array of variables.
        Returns:
            Value of the polynomial at c.
        """

        if self._array_coefficients is not None and self._constant_term is not None:
            monomials = self.compute_monomials(c)
            return self._constant_term + _numpy_multi_dispatch(
                self._array_coefficients, monomials, path="tensordot", axes=(0, 0)
            )
        elif self._array_coefficients is not None:
            monomials = self.compute_monomials(c)
            return _numpy_multi_dispatch(
                self._array_coefficients, monomials, path="tensordot", axes=(0, 0)
            )
        else:
            return self._constant_term


def _get_monomial_compute_function(multisets: List[Multiset]) -> Callable:
    """Construct a vectorized function for computing multivariable monomial terms indicated by
    multisets.

    The returned function takes in the individual variables as an array, and returns an array of
    computed monomial terms in the order indicated by multisets.

    The returned function is vectorized in the sense that the supplied first order terms can be
    arrays.

    The algorithm computes monomial terms of increasing order, recursively utilizing lower order
    terms.

    Args:
        multisets: list of multisets.

    Returns:
        Callable: Vectorized function for computing monomials.
    """
    if multisets is None or len(multisets) == 0:
        return lambda c: None

    complete_multiset_list = _get_all_submultisets(multisets)

    (
        first_order_terms,
        first_order_range,
        left_indices,
        right_indices,
        update_ranges,
    ) = _get_recursive_monomial_rule(complete_multiset_list)

    multiset_len = len(complete_multiset_list)

    location_list = np.array(
        [complete_multiset_list.index(multiset) for multiset in multisets], dtype=int
    )

    def monomial_function(c):
        shape = [multiset_len] + list(c.shape[1:])
        mono_vec = np.empty(shape=shape, dtype=complex)
        mono_vec[first_order_range[0] : first_order_range[1]] = c[first_order_terms]

        for left_index, right_index, update_range in zip(
            left_indices, right_indices, update_ranges
        ):
            mono_vec[update_range[0] : update_range[1]] = (
                mono_vec[left_index] * mono_vec[right_index]
            )

        return mono_vec[location_list]

    return monomial_function


# pylint: disable=invalid-name
def _get_monomial_compute_function_jax(multisets: List) -> Callable:
    """JAX version of _get_monomial_compute_function."""

    if multisets is None or len(multisets) == 0:
        return lambda c: None

    complete_multiset_list = _get_all_submultisets(multisets)

    first_order_terms, _, left_indices, right_indices, _ = _get_recursive_monomial_rule(
        complete_multiset_list
    )
    location_list = np.array(
        [complete_multiset_list.index(multiset) for multiset in multisets], dtype=int
    )

    # initial function sets up required first order terms
    def monomial_function_init(c):
        return c[first_order_terms]

    monomial_function = monomial_function_init

    # function for generating next update and compositions to avoid looping reference issues
    def get_next_update_func(left_index, right_index):
        def update_next(mono_vec):
            return jnp.append(mono_vec, mono_vec[left_index] * mono_vec[right_index], axis=0)

        return update_next

    def compose_functions(f, g):
        def new_func(x):
            return g(f(x))

        return new_func

    # recursively compose updates with monomial_function
    for left_index, right_index in zip(left_indices, right_indices):
        next_update = get_next_update_func(left_index, right_index)
        monomial_function = compose_functions(monomial_function, next_update)

    # return only the requested terms
    def trimmed_output_function(c):
        return monomial_function(c)[location_list]

    return trimmed_output_function


def _get_recursive_monomial_rule(complete_multisets: List) -> Tuple:
    """Helper function for _get_monomial_compute_function and _get_monomial_compute_function_jax;
    computes a representation of the algorithm for computing monomials that is used by both
    functions.

    complete_multisets is assumed to be closed under taking submultisets and in canonical order.

    Args:
        complete_multisets: Description of monomial terms.

    Returns:
        Tuple: Collection of lists organizing computation for both
        _get_monomial_compute_function and _get_monomial_compute_function_jax.
    """

    # first, construct representation of recursive rule explicitly in terms of multisets
    first_order_terms = []
    left_terms = []
    right_terms = []
    current_left = -1
    current_right_list = []
    current_len = 2

    # convert multisets to list representation
    complete_multisets = [_multiset_to_sorted_list(multiset) for multiset in complete_multisets]

    for multiset in complete_multisets:
        if len(multiset) == 1:
            first_order_terms.append(multiset[0])
        else:
            if multiset[0] != current_left or len(multiset) != current_len:
                current_len = len(multiset)
                if current_left != -1:
                    left_terms.append(current_left)
                    right_terms.append(current_right_list)

                current_left = multiset[0]
                current_right_list = [multiset[1:]]
            else:
                current_right_list.append(multiset[1:])

    # if current_left is still -1, then only first order terms exist
    if current_left == -1:
        return np.array(first_order_terms), [0, len(first_order_terms)], [], [], []

    # proceed assuming at least one term above first order exists
    # append the last one
    left_terms.append(current_left)
    right_terms.append(current_right_list)

    # convert representation of rule in terms of multisets into one in terms of
    # array indices

    # set up arrays
    first_order_terms = np.array(first_order_terms, dtype=int)

    left_indices = []
    right_indices = []
    for left_term, right_term in zip(left_terms, right_terms):
        left_indices.append(complete_multisets.index([left_term]))

        right_index_list = []
        for term in right_term:
            right_index_list.append(complete_multisets.index(term))

        right_indices.append(np.array(right_index_list, dtype=int))

    # set up index updating ranges
    first_order_range = [0, len(first_order_terms)]
    update_ranges = []
    current_idx = first_order_range[1]
    for right_index in right_indices:
        next_idx = current_idx + len(right_index)
        update_ranges.append([current_idx, next_idx])
        current_idx = next_idx

    return (
        np.array(first_order_terms),
        first_order_range,
        np.array(left_indices),
        right_indices,
        update_ranges,
    )


def _array_polynomial_distributive_binary_op(
    ap1: ArrayPolynomial,
    ap2: ArrayPolynomial,
    binary_op: Callable,
    monomial_filter: Optional[Callable] = None,
) -> ArrayPolynomial:
    """Apply a distributive binary op on two array polynomials."""

    # determine list of Multisets required for monomial labels, including filtering
    all_multisets = []

    # if no filter is provided, set to always return True
    if monomial_filter is None:

        def monomial_filter(_):
            return True

    if ap1.constant_term is not None:
        for multiset in ap2.monomial_labels:
            if monomial_filter(multiset) and multiset not in all_multisets:
                all_multisets.append(multiset)
    if ap2.constant_term is not None:
        for multiset in ap1.monomial_labels:
            if monomial_filter(multiset) and multiset not in all_multisets:
                all_multisets.append(multiset)

    for I, J in product(ap1.monomial_labels, ap2.monomial_labels):
        IuJ = I + J
        if monomial_filter(IuJ) and IuJ not in all_multisets:
            all_multisets.append(IuJ)

    all_multisets = _sorted_multisets(all_multisets)

    # setup constant term
    new_constant_term = None
    if (
        ap1.constant_term is not None
        and ap2.constant_term is not None
        and monomial_filter(Multiset({}))
    ):
        new_constant_term = binary_op(ap1.constant_term, ap2.constant_term)

    # return constant case
    if not all_multisets:
        return ArrayPolynomial(constant_term=new_constant_term)

    # iteratively construct custom multiplication rule,
    # temporarily treating the constant terms as index -1
    operation_rule = []
    for multiset in all_multisets:
        rule_indices = []

        if multiset in ap1.monomial_labels:
            idx = ap1.monomial_labels.index(multiset)
            rule_indices.append([idx, -1])

        if multiset in ap2.monomial_labels:
            idx = ap2.monomial_labels.index(multiset)
            rule_indices.append([-1, idx])

        if len(multiset) > 1:
            for I, J in zip(*_submultisets_and_complements(multiset)):
                if I in ap1.monomial_labels and J in ap2.monomial_labels:
                    rule_indices.append(
                        [ap1.monomial_labels.index(I), ap2.monomial_labels.index(J)]
                    )

        # if non-empty,
        if rule_indices:
            operation_rule.append((np.ones(len(rule_indices)), np.array(rule_indices)))

    lmats = None
    if ap1.constant_term is not None:
        lmats = unp.expand_dims(ap1.constant_term, 0)
    else:
        lmats = unp.expand_dims(unp.zeros_like(ap1.array_coefficients[0]), 0)

    if ap1.array_coefficients is not None:
        lmats = _numpy_multi_dispatch(lmats, ap1.array_coefficients, path="append", axis=0)

    rmats = None
    if ap2.constant_term is not None:
        rmats = unp.expand_dims(ap2.constant_term, 0)
    else:
        rmats = unp.expand_dims(unp.zeros_like(ap2.array_coefficients[0]), 0)

    if ap2.array_coefficients is not None:
        rmats = _numpy_multi_dispatch(rmats, ap2.array_coefficients, path="append", axis=0)

    custom_binary_op = _CustomBinaryOp(
        operation_rule=operation_rule,
        binary_op=binary_op,
        index_offset=1,
    )
    new_array_coefficients = custom_binary_op(lmats, rmats)

    return ArrayPolynomial(
        array_coefficients=new_array_coefficients,
        monomial_labels=all_multisets,
        constant_term=new_constant_term,
    )


def _array_polynomial_addition(
    ap1: ArrayPolynomial,
    ap2: ArrayPolynomial,
    monomial_filter: Optional[Callable] = None,
) -> ArrayPolynomial:
    """Add two ArrayPolynomials."""

    for a, b in zip(ap1.shape[::-1], ap2.shape[::-1]):
        if not (a == 1 or b == 1 or a == b):
            raise QiskitError(
                "ArrayPolynomial addition requires shapes be broadcastable to eachother."
            )

    if monomial_filter is None:

        def monomial_filter(_):
            return True

    # construct constant term
    new_constant_term = None

    # if constant term is to be included, determine what it is
    if monomial_filter(Multiset({})):
        if ap1.constant_term is not None and ap2.constant_term is not None:
            new_constant_term = ap1.constant_term + ap2.constant_term
        elif ap1.constant_term is not None:
            new_constant_term = ap1.constant_term
        elif ap2.constant_term is not None:
            new_constant_term = ap2.constant_term

    # exit early if both polynomials are constant
    if ap1.array_coefficients is None and ap2.array_coefficients is None:
        return ArrayPolynomial(constant_term=new_constant_term)

    # construct list of admissable multisets and sort into canonical order
    new_multisets = []
    for multiset in ap1.monomial_labels + ap2.monomial_labels:
        if monomial_filter(multiset) and multiset not in new_multisets:
            new_multisets.append(multiset)
    new_multisets = _sorted_multisets(new_multisets)

    # construct index order mapping for each polynomial
    idx1 = []
    idx2 = []
    for multiset in new_multisets:
        if multiset in ap1.monomial_labels:
            idx1.append(ap1.monomial_labels.index(multiset))
        else:
            idx1.append(-1)

        if multiset in ap2.monomial_labels:
            idx2.append(ap2.monomial_labels.index(multiset))
        else:
            idx2.append(-1)

    # if either is empty, pad with single -1, then convert to array
    idx1 = idx1 or [-1]
    idx2 = idx2 or [-1]
    idx1 = np.array(idx1)
    idx2 = np.array(idx2)

    # append zero to the coefficient arrays
    array_coefficients1 = np.zeros((1,) + ap1.shape, dtype=complex)
    array_coefficients2 = np.zeros((1,) + ap1.shape, dtype=complex)
    if ap1.array_coefficients is not None:
        array_coefficients1 = unp.append(ap1.array_coefficients, array_coefficients1, axis=0)
    if ap2.array_coefficients is not None:
        array_coefficients2 = unp.append(ap2.array_coefficients, array_coefficients2, axis=0)

    new_coefficients = array_coefficients1[idx1] + array_coefficients2[idx2]

    return ArrayPolynomial(
        array_coefficients=new_coefficients,
        monomial_labels=new_multisets,
        constant_term=new_constant_term,
    )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

r"""
Core functionality for computing Dyson series [1] and Magnus expansion [2, 3] terms. Specifically,
Dyson-like terms are computed via the algorithm in [4], and Dyson series and Magnus expansion terms
are computed via the method in [5].

References:
    1. F. Dyson, *The radiation theories of Tomonaga, Schwinger, and Feynman*,
       Phys. Rev. 75, 486-502
    2. W. Magnus, *On the exponential solution of differential equations*
       *for a linear operator*, Commun. Pure Appl. Math. 7, 649-73
    3. S. Blanes, F. Casas, J. Oteo, J. Ros, *The Magnus expansion and some*
       *of its applications*, Phys. Rep. 470, 151-238
    4. H. Haas, D. Puzzuoli, F. Zhang, D. Cory, *Engineering Effective Hamiltonians*,
       New J. Phys. 21, 103011 (2019).
    5. D. Puzzuoli, S. F. Lin, M. Malekakhlagh, E. Pritchett, B. Rosand, C. J. Wood, *Algorithms
    for perturbative analysis and simulation of quantum dynamics*, arXiv:2210.11595 (2022)
"""

from typing import Optional, List, Callable, Tuple, Union

import numpy as np
from scipy.special import factorial

from scipy.integrate._ivp.ivp import OdeResult

from multiset import Multiset

from qiskit_dynamics import ArrayLike
from qiskit_dynamics.solvers.solver_functions import solve_ode

from qiskit_dynamics.perturbation.custom_binary_op import (
    _compile_custom_operation_rule,
    _CustomMatmul,
)

from qiskit_dynamics.perturbation.multiset_utils import (
    _get_all_submultisets,
    _submultiset_filter,
    _submultisets_and_complements,
)

from .perturbation_data import PowerSeriesData, DysonLikeData

try:
    import jax.numpy as jnp
    from jax.lax import scan
    from jax import vmap
except ImportError:
    pass


def _solve_lmde_dyson(
    perturbations: List[Callable],
    t_span: ArrayLike,
    dyson_terms: Union[List[List[int]], List[Multiset]],
    perturbation_labels: Optional[Union[List[List], List[Multiset]]] = None,
    generator: Optional[Callable] = None,
    y0: Optional[ArrayLike] = None,
    dyson_in_frame: Optional[bool] = True,
    dyson_like: Optional[bool] = False,
    integration_method: Optional[str] = "DOP853",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
) -> OdeResult:
    """Helper function for computing Dyson terms using methods in References [4, 5]. See
    documentation for :meth:`solve_lmde_perturbation`.

    Args:
        perturbations: List of callable matrix functions to appear in Dyson terms.
        t_span: Integration limits.
        dyson_terms: Terms to compute.
        perturbation_labels: Ordering/specification of the elements of perturbations. Only used for
            dyson_like==False.
        generator: Optional frame generator.
        y0: Optional initial state for frame generator LMDE.
        dyson_in_frame: Whether to return the Dyson terms in the frame of the the frame generator.
        dyson_like: Compute either Dyson or Dyson-like terms.
        integration_method: Integration method.
        t_eval: Optional additional time points at which to return the solution.
        **kwargs: Additional arguments to pass to the solver.

    Returns:
        OdeResult
    """
    mat_dim = perturbations[0](t_span[0]).shape[0]

    if generator is None:
        # pylint: disable=unused-argument
        def default_generator(t):
            return np.zeros((mat_dim, mat_dim), dtype=complex)

        generator = default_generator

    if y0 is None:
        y0 = np.eye(mat_dim, dtype=complex)

    # construct term list depending on whether computation is dyson or dyson_like
    complete_term_list = None
    if dyson_like:
        complete_term_list = _get_complete_dyson_like_indices(dyson_terms)
    else:
        complete_term_list = _get_all_submultisets(dyson_terms)

    dyson_rhs = _setup_dyson_rhs(
        generator,
        perturbations,
        complete_term_list,
        mat_dim,
        dyson_like=dyson_like,
        perturbation_labels=perturbation_labels,
    )

    # initial state
    y0 = np.append(
        np.expand_dims(y0, 0),
        np.zeros((len(complete_term_list), y0.shape[-2], y0.shape[-1]), dtype=complex),
        axis=0,
    )

    results = solve_ode(
        rhs=dyson_rhs, t_span=t_span, y0=y0, method=integration_method, t_eval=t_eval, **kwargs
    )

    # extract Dyson terms and the solution to the base LMDE
    results.y = results.y.transpose((1, 0, 2, 3))
    dyson_terms = results.y[1:]
    results.y = results.y[0]

    if dyson_in_frame:
        for idx, dyson_term in enumerate(dyson_terms):
            dyson_terms[idx] = np.linalg.solve(results.y, dyson_term)

    expansion_type = "dyson"
    DataClass = PowerSeriesData
    if dyson_like:
        expansion_type = "dyson_like"
        DataClass = DysonLikeData

    results.perturbation_data = DataClass(
        data=dyson_terms,
        labels=complete_term_list,
        metadata={"expansion_type": expansion_type},
    )

    return results


def _solve_lmde_magnus(
    perturbations: List[Callable],
    t_span: ArrayLike,
    magnus_terms: List[Multiset],
    perturbation_labels: Optional[List[Multiset]] = None,
    generator: Optional[Callable] = None,
    y0: Optional[ArrayLike] = None,
    integration_method: Optional[str] = "DOP853",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
) -> OdeResult:
    """Helper function for computing Magnus terms using method in Reference [5]. See documentation
    for :meth:`solve_lmde_perturbation`.

    Args:
        perturbations: List of callable matrix functions to appear in Dyson terms.
        t_span: Integration limits.
        magnus_terms: Terms to compute.
        perturbation_labels: Ordering/specification of the elements of perturbations.
        generator: Optional frame generator.
        y0: Optional initial state for frame generator LMDE.
        integration_method: Integration method.
        t_eval: Optional additional time points at which to return the solution.
        **kwargs: Additional arguments to pass to the solver.

    Returns:
        OdeResult
    """

    # first compute Dyson terms
    results = _solve_lmde_dyson(
        perturbations,
        t_span,
        dyson_terms=magnus_terms,
        perturbation_labels=perturbation_labels,
        generator=generator,
        y0=y0,
        dyson_in_frame=True,
        dyson_like=False,
        integration_method=integration_method,
        t_eval=t_eval,
        **kwargs,
    )

    # compute Magnus terms from Dyson and update the results
    magnus_terms = _magnus_from_dyson(
        results.perturbation_data.labels, results.perturbation_data.data
    )
    results.perturbation_data.metadata = {"expansion_type": "magnus"}
    results.perturbation_data.data = magnus_terms

    return results


def _solve_lmde_dyson_jax(
    perturbations: List[Callable],
    t_span: ArrayLike,
    dyson_terms: Union[List[List[int]], List[Multiset]],
    perturbation_labels: Optional[Union[List[List], List[Multiset]]] = None,
    generator: Optional[Callable] = None,
    y0: Optional[ArrayLike] = None,
    dyson_in_frame: Optional[bool] = True,
    dyson_like: Optional[bool] = True,
    integration_method: Optional[str] = "jax_odeint",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
) -> OdeResult:
    """JAX version of ``_solve_lmde_dyson``. See documentation for :meth:`solve_lmde_perturbation`.

    Args:
        perturbations: List of callable matrix functions to appear in Dyson terms.
        t_span: Integration limits.
        dyson_terms: Terms to compute.
        perturbation_labels: Ordering/specification of the elements of perturbations. Only used if
            dyson_like==False.
        generator: Optional frame generator.
        y0: Optional initial state for frame generator LMDE.
        dyson_in_frame: Whether to return the Dyson terms in the frame of the the frame generator.
        dyson_like: Compute either Dyson or Dyson-like terms.
        integration_method: Integration method.
        t_eval: Optional additional time points at which to return the solution.
        **kwargs: Additional arguments to pass to the solver.

    Returns:
        OdeResult
    """

    mat_dim = perturbations[0](t_span[0]).shape[0]

    if generator is None:
        # pylint: disable=unused-argument
        def default_generator(t):
            return jnp.zeros((mat_dim, mat_dim), dtype=complex)

        generator = default_generator

    if y0 is None:
        y0 = jnp.eye(mat_dim, dtype=complex)

    # construct term list an RHS based on whether dyson or dyson-like
    complete_term_list = None
    if dyson_like:
        complete_term_list = _get_complete_dyson_like_indices(dyson_terms)
    else:
        complete_term_list = _get_all_submultisets(dyson_terms)

    dyson_rhs = _setup_dyson_rhs_jax(
        generator,
        perturbations,
        complete_term_list,
        dyson_like=dyson_like,
        perturbation_labels=perturbation_labels,
    )

    # initial state
    y0 = jnp.append(
        jnp.expand_dims(y0, 0),
        jnp.zeros((len(complete_term_list), y0.shape[-2], y0.shape[-1]), dtype=complex),
        axis=0,
    )

    results = solve_ode(
        rhs=dyson_rhs, t_span=t_span, y0=y0, method=integration_method, t_eval=t_eval, **kwargs
    )

    # extract Dyson terms and the solution to the base LMDE
    results.y = results.y.transpose((1, 0, 2, 3))
    dyson_terms = jnp.array(results.y[1:])
    results.y = jnp.array(results.y[0])

    if dyson_in_frame:
        dyson_terms = vmap(lambda x: jnp.linalg.solve(results.y, x))(dyson_terms)

    expansion_type = "dyson"
    DataClass = PowerSeriesData
    if dyson_like:
        expansion_type = "dyson_like"
        DataClass = DysonLikeData

    results.perturbation_data = DataClass(
        data=dyson_terms,
        labels=complete_term_list,
        metadata={"expansion_type": expansion_type},
    )

    return results


def _solve_lmde_magnus_jax(
    perturbations: List[Callable],
    t_span: ArrayLike,
    magnus_terms: List[Multiset],
    perturbation_labels: Optional[List[Multiset]] = None,
    generator: Optional[Callable] = None,
    y0: Optional[ArrayLike] = None,
    integration_method: Optional[str] = "DOP853",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
) -> OdeResult:
    """JAX version of ``_solve_lmde_magnus``. See documentation for :meth:`solve_lmde_perturbation`.

    Args:
        perturbations: List of callable matrix functions to appear in Dyson terms.
        t_span: Integration limits.
        magnus_terms: Terms to compute.
        perturbation_labels: Ordering/specification of the elements of perturbations.
        generator: Optional frame generator.
        y0: Optional initial state for frame generator LMDE.
        integration_method: Integration method.
        t_eval: Optional additional time points at which to return the solution.
        **kwargs: Additional arguments to pass to the solver.

    Returns:
        OdeResult
    """

    # first compute Dyson terms
    results = _solve_lmde_dyson_jax(
        perturbations,
        t_span,
        dyson_terms=magnus_terms,
        perturbation_labels=perturbation_labels,
        generator=generator,
        y0=y0,
        dyson_in_frame=True,
        dyson_like=False,
        integration_method=integration_method,
        t_eval=t_eval,
        **kwargs,
    )

    # compute Magnus terms from Dyson and update the results
    magnus_terms = _magnus_from_dyson_jax(
        results.perturbation_data.labels, results.perturbation_data.data
    )
    results.perturbation_data.metadata = {"expansion_type": "magnus"}
    results.perturbation_data.data = magnus_terms

    return results


def _setup_dyson_rhs(
    generator: Callable,
    perturbations: List[Callable],
    oc_dyson_indices: List[Multiset],
    mat_dim: int,
    dyson_like: Optional[bool] = True,
    perturbation_labels: Optional[List[Multiset]] = None,
) -> Callable:
    """Construct the RHS function for propagating Dyson terms.

    Args:
        generator: The frame generator G.
        perturbations: List of matrix functions appearing in Dyson terms.
        oc_dyson_indices: Ordered complete list of Dyson terms to compute.
        mat_dim: Dimension of outputs of generator and functions in perturbations.
        dyson_like: Whether the computation is for Dyson or Dyson-like terms.
        perturbation_labels: List of lists specifying index information for perturbations. Only used
            when dyson_like==False.

    Returns:
        Callable
    """
    lmult_rule = None
    perturbations_evaluation_order = None
    if dyson_like:
        generator_eval_indices = _required_dyson_generator_indices(oc_dyson_indices)
        perturbations_evaluation_order = [0] + [idx + 1 for idx in generator_eval_indices]
        lmult_rule = _get_dyson_like_lmult_rule(oc_dyson_indices, generator_eval_indices)
    else:
        # filter members of perturbations required for given list of dyson terms
        if perturbation_labels is None:
            perturbation_labels = [[idx] for idx in range(len(perturbations))]
        reduced_perturbation_labels = _submultiset_filter(perturbation_labels, oc_dyson_indices)
        perturbations_evaluation_order = [0] + [
            perturbation_labels.index(multiset) + 1 for multiset in reduced_perturbation_labels
        ]
        lmult_rule = _get_dyson_lmult_rule(oc_dyson_indices, reduced_perturbation_labels)

    custom_matmul = _CustomMatmul(lmult_rule, index_offset=1)

    perturbations_evaluate_len = len(perturbations_evaluation_order)
    new_list = [generator] + perturbations

    def gen_evaluator(t):
        mat = np.empty((perturbations_evaluate_len, mat_dim, mat_dim), dtype=complex)

        for idx, a_idx in enumerate(perturbations_evaluation_order):
            mat[idx] = new_list[a_idx](t)

        return mat

    def dyson_rhs(t, y):
        return custom_matmul(gen_evaluator(t), y)

    return dyson_rhs


def _setup_dyson_rhs_jax(
    generator: Callable,
    perturbations: List[Callable],
    oc_dyson_indices: List[Multiset],
    dyson_like: Optional[bool] = True,
    perturbation_labels: Optional[List[Multiset]] = None,
) -> Callable:
    """JAX version of _setup_dyson_rhs.

    Args:
        generator: The frame generator G.
        perturbations: List of matrix functions appearing in Dyson terms.
        oc_dyson_indices: Ordered complete list of Dyson terms to compute.
        dyson_like: Whether the computation is for Dyson or Dyson-like terms.
        perturbation_labels: List of lists specifying index information for perturbations. Only used
            when dyson_like==False.

    Returns:
        Callable
    """

    lmult_rule = None
    perturbations_evaluation_order = None
    if dyson_like:
        generator_eval_indices = _required_dyson_generator_indices(oc_dyson_indices)
        perturbations_evaluation_order = [0] + [idx + 1 for idx in generator_eval_indices]
        lmult_rule = _get_dyson_like_lmult_rule(oc_dyson_indices, generator_eval_indices)
    else:
        # filter members of perturbations required for given list of dyson terms
        if perturbation_labels is None:
            perturbation_labels = [[idx] for idx in range(len(perturbations))]
        reduced_perturbation_labels = _submultiset_filter(perturbation_labels, oc_dyson_indices)
        perturbations_evaluation_order = [0] + [
            perturbation_labels.index(multiset) + 1 for multiset in reduced_perturbation_labels
        ]
        lmult_rule = _get_dyson_lmult_rule(oc_dyson_indices, reduced_perturbation_labels)

    custom_matmul = _CustomMatmul(lmult_rule, index_offset=1)

    perturbations_evaluation_order = np.array(perturbations_evaluation_order, dtype=int)

    new_list = [generator] + perturbations

    def multiple_eval(t):
        return jnp.array([new_list[idx](t) for idx in perturbations_evaluation_order])

    def dyson_rhs(t, y):
        return custom_matmul(multiple_eval(t), y)

    return dyson_rhs


def _required_dyson_generator_indices(complete_dyson_indices: List) -> List:
    """Given a complete list of dyson indices, determine which generator terms are required."""
    generator_indices = []
    for term in complete_dyson_indices:
        if term[0] not in generator_indices:
            generator_indices.append(term[0])

    generator_indices.sort()
    return generator_indices


def _get_dyson_like_lmult_rule(
    complete_dyson_indices: List[List[int]], generator_indices: List[List[int]]
) -> List:
    """Construct custom product rules, in the format required by ``custom_product``, for a given set
    of Dyson terms.

    Assumption: the supplied list is complete, i.e. if a term depends on other terms, then the terms
    it depends on are also in the list.

    Convention: G(t) is given the index -1 to preserve the indexing of perturbations.

    Args:
        complete_dyson_indices: Complete list of Dyson terms.
        generator_indices: List of required generator terms.

    Returns:
        List: lmult rule.
    """

    # construct multiplication rules
    lmult_rule = [(np.array([1.0]), np.array([[-1, -1]]))]

    for term_idx, term in enumerate(complete_dyson_indices):
        if len(term) == 1:
            l_idx = generator_indices.index(term[0])
            lmult_rule.append((np.array([1.0, 1.0]), np.array([[-1, term_idx], [l_idx, -1]])))
        else:
            # self multiplied by generator
            lmult_indices = [[-1, term_idx]]
            # the left index is the first entry in term
            # check if it is required before adding

            l_idx = generator_indices.index(term[0])
            r_idx = complete_dyson_indices.index(term[1:])
            lmult_indices.append([l_idx, r_idx])

            lmult_rule.append(
                (np.ones(len(lmult_indices), dtype=float), np.array(lmult_indices, dtype=int))
            )

    return lmult_rule


def _get_complete_dyson_like_indices(dyson_terms: List[List[int]]) -> List[List[int]]:
    """Given a list of Dyson terms to compute specified as lists of indices, recursively construct
    all other Dyson terms that need to be computed, returned as a list, ordered by increasing Dyson
    order, and in lexicographic order within an order.

    Args:
        dyson_terms: Terms to compute.

    Returns:
        list: List of all terms that need to be computed.
    """

    max_order = max(map(len, dyson_terms))
    term_dict = {k: [] for k in range(1, max_order + 1)}

    # first populate with requested terms
    for term in dyson_terms:
        order = len(term)
        if term not in term_dict[order]:
            term_dict[order].append(list(term))

    # loop through orders in reverse order
    for order in range(max_order, 1, -1):
        for term in term_dict[order]:
            term = list(term)
            if term[1:] not in term_dict[order - 1]:
                term_dict[order - 1].append(term[1:])

    ordered_term_list = []

    for order in range(1, max(term_dict.keys()) + 1):
        ordered_term_list += term_dict[order]

    # sort in terms of increasing length and lexicographic order
    ordered_term_list.sort(key=str)
    ordered_term_list.sort(key=len)

    return ordered_term_list


def _magnus_from_dyson(complete_index_multisets: List[Multiset], dyson_terms: np.array) -> np.array:
    """Compute magnus terms from dyson terms using the recursion relation presented in [5]. The term
    "Q Matrices" in helper functions refers to the matrices used in the recursion relation in [5].

    Args:
        complete_index_multisets: A complete and canonically ordered list of index Multisets.
        dyson_terms: Array of Dyson terms.

    Returns:
        np.array: The Magnus terms.
    """

    ordered_q_terms = _get_q_term_list(complete_index_multisets)
    start_idx, magnus_indices, stacked_q_update_rules = _q_recursive_compiled_rules(ordered_q_terms)

    # if all terms are first order, nothing needs to be done
    if start_idx == len(dyson_terms):
        return dyson_terms

    # initialize array of q matrices with dyson terms
    q_mat_shape = (len(ordered_q_terms) + 1,) + dyson_terms.shape[1:]
    q_mat = np.zeros(q_mat_shape, dtype=complex)
    # add an identity at the end
    q_mat[-1] = np.broadcast_to(np.eye(q_mat.shape[-1], dtype=complex), q_mat.shape[1:])
    q_mat[magnus_indices] = dyson_terms

    index_list = start_idx + np.arange(len(stacked_q_update_rules[0]))

    for rule_idx, mat_idx in enumerate(index_list):
        compiled_rule = (
            stacked_q_update_rules[0][rule_idx],
            (stacked_q_update_rules[1][0][rule_idx], stacked_q_update_rules[1][1][rule_idx]),
        )
        custom_matmul = _CustomMatmul(compiled_rule, operation_rule_compiled=True)
        q_mat[mat_idx] = custom_matmul(q_mat, q_mat)[0]

    return q_mat[magnus_indices]


def _magnus_from_dyson_jax(
    complete_index_multisets: List[Multiset], dyson_terms: np.array
) -> np.array:
    """JAX version of _magnus_from_dyson."""

    ordered_q_terms = _get_q_term_list(complete_index_multisets)
    start_idx, magnus_indices, stacked_q_update_rules = _q_recursive_compiled_rules(ordered_q_terms)

    # if all terms are first order, nothing needs to be done
    if start_idx == len(dyson_terms):
        return dyson_terms

    # initialize array of q matrices with dyson terms
    q_mat_shape = (len(ordered_q_terms) + 1,) + dyson_terms.shape[1:]
    q_init = jnp.zeros(q_mat_shape, dtype=complex)
    q_init = q_init.at[magnus_indices].set(dyson_terms)
    q_init = q_init.at[-1].set(
        jnp.broadcast_to(jnp.eye(q_init.shape[-1], dtype=complex), q_init.shape[1:])
    )

    index_list = start_idx + jnp.arange(len(stacked_q_update_rules[0]))

    def scan_fun(B, x):
        idx, compiled_rule = x
        custom_matmul = _CustomMatmul(compiled_rule, operation_rule_compiled=True)
        update = custom_matmul(B, B)[0]
        new_B = B.at[idx].set(update)
        return new_B, None

    q_mats = scan(scan_fun, init=q_init, xs=(index_list, stacked_q_update_rules))[0]

    return q_mats[magnus_indices]


def _q_recursive_compiled_rules(ordered_q_terms: List) -> Tuple[int, np.array, Tuple]:
    """Construct compiled custom product rules for recursive computation of Q matrices.

    Note: this function "stacks" the rules into a single tuple whose formatting is chosen to be
    usable with jax loop constructs.

    Args:
        ordered_q_terms: Ordered list of Q matrix specifications.

    Returns:
        start_idx: the index at q_terms need to start being updated
        magnus_indices: The locations in the q matrix list corresponding to Magnus terms.
        stacked_compiled_rules: stacked rules as per the above note
    """

    # create list of locations in q_term_list corresponding to Magnus terms
    # and find start index
    start_idx = 0
    magnus_indices = []
    for idx, q_term in enumerate(ordered_q_terms):
        if q_term[1] == 1:
            magnus_indices.append(idx)

        if len(q_term[0]) == 1:
            start_idx += 1

    magnus_indices = np.array(magnus_indices)

    # first, construct rules, and determine a maximum length
    max_unique_mults = 0
    max_linear_rule = 0
    rules = []
    for q_term in ordered_q_terms[start_idx:]:
        rule = _q_product_rule(q_term, ordered_q_terms)
        rules.append(rule)

        unique_mults, linear_rule = _compile_custom_operation_rule(rule)

        max_unique_mults = max(max_unique_mults, len(unique_mults))
        max_linear_rule = max(max_linear_rule, linear_rule[0].shape[1])

    stacked_unique_mults = []
    stacked_linear_rules = ([], [])
    for rule in rules:
        unique_mults, linear_rule = _compile_custom_operation_rule(
            rule, unique_evaluation_len=max_unique_mults, linear_combo_len=max_linear_rule
        )
        stacked_unique_mults.append(unique_mults)
        stacked_linear_rules[0].append(linear_rule[0])
        stacked_linear_rules[1].append(linear_rule[1])

    # convert to arrays and put into standard format
    stacked_unique_mults = np.array(stacked_unique_mults)
    stacked_linear_combo_rule = (
        np.array(stacked_linear_rules[0]),
        np.array(np.array(stacked_linear_rules[1])),
    )
    stacked_compiled_rules = (stacked_unique_mults, stacked_linear_combo_rule)

    return start_idx, magnus_indices, stacked_compiled_rules


def _q_product_rule(q_term: Tuple, oc_q_term_list: List[Tuple]) -> List:
    """Given a specification of a Q matrix and an ordered complete list of Q matrix specifications,
    constructs the recursion relation required to compute q_term, specified as a custom product rule
    for instantiating a CustomProduct.

    Note:
        - This assumes len(sym_index) > 1, as the purpose of this function is to apply the recursion
          rules, and no rule is required when len(sym_index) == 1.
        - This function also assumes that q_term, and oc_q_term_list are correctly formatted in
          terms of internal sorting.

    Args:
        q_term: Tuple with a Multiset and a product order (int)
        oc_q_term_list: Ordered complete list of a q terms.

    Returns:
        List
    """

    sym_index, q_term_order = q_term
    q_term_idx = oc_q_term_list.index(q_term)
    q_term_len = len(sym_index)

    if q_term_order == 1:
        # if the order is 1, it is just a linear combination of lower terms
        coeffs = np.append(1.0, -1 / factorial(range(2, q_term_len + 1), exact=True))

        products = [[len(oc_q_term_list), q_term_idx]]
        for prod_order in range(2, q_term_len + 1):
            products.append([len(oc_q_term_list), oc_q_term_list.index((sym_index, prod_order))])

        return [(coeffs, np.array(products))]
    else:
        # construct a list of products
        # need to consider all possible sub-multisets of the Multiset index in q_term
        products = []
        submultisets, complements = _submultisets_and_complements(
            sym_index, len(sym_index) - (q_term_order - 1) + 1
        )

        for subset, complement in zip(submultisets, complements):
            product = [
                oc_q_term_list.index((subset, 1)),
                oc_q_term_list.index((complement, q_term_order - 1)),
            ]
            if product not in products:
                products.append(product)

        coeffs = np.ones(len(products), dtype=float)
        return [(coeffs, np.array(products))]


def _get_q_term_list(complete_index_multisets: List[Multiset]) -> List:
    """Construct a specification of the recursive Q matrices required to compute all Magnus terms
    specified by ``complete_index_multisets``. Each Q matrix is specified as a 2-tuple with first
    entry a Multiset indexing the term, and second entry the product order of the Q matrix.

    Note: This function assumes ``complete_index_multisets`` are canonically ordered and correctly
    formatted. The output is then a canonical ordering of the Q matrices.

    Args:
        complete_index_multisets: Canonically ordered complete list of index Multisets.

    Returns:
        List: Q matrix specification
    """

    q_terms = []
    for term in complete_index_multisets:
        for order in range(len(term), 0, -1):
            q_terms.append((term, order))

    return q_terms


def _get_dyson_lmult_rule(
    complete_index_multisets: List[Multiset], perturbation_labels: Optional[List[Multiset]] = None
) -> List:
    """Given a complete list of index multisets, return the lmult rule in the format required for
    ``CustomProduct``. Note, the generator :math:`G(t)` is encoded as index ``-1``, as it will be
    prepended to the list of A matrices. Similarly the index of the solution to the LMDE with
    generator :math:`G(t)` is encoded as ``-1``.

    While not required within the logic of this function, the input should be canonically ordered
    according to ``_get_all_submultisets``.

    Args:
        complete_index_multisets: List of complete multisets.
        perturbation_labels: List of index multisets describing perturbations.

    Returns:
        List: Left multiplication rule description.
    """

    # If perturbation_labels is not specified, use the elements of complete_index_multisets
    # of length 1
    if perturbation_labels is None:
        perturbation_labels = []
        for entry in complete_index_multisets:
            if len(entry) == 1:
                perturbation_labels.append(entry)
            else:
                break

    # construct multiplication rules
    lmult_rule = [(np.array([1.0]), np.array([[-1, -1]]))]

    for term_idx, term in enumerate(complete_index_multisets):
        if len(term) == 1 and term in perturbation_labels:
            # if term not in perturbation_labels for len(term) == 1 the corresponding Dyson integral
            # will always be 0
            lmult_rule.append(
                (
                    np.array([1.0, 1.0]),
                    np.array([[-1, term_idx], [perturbation_labels.index(term), -1]]),
                )
            )
        else:
            # self multiplied by base generator
            lmult_indices = [[-1, term_idx]]

            for l_idx, l_term in enumerate(perturbation_labels):
                if l_term <= term:
                    if len(l_term) == len(term):
                        lmult_indices.append([l_idx, -1])
                    else:
                        r_term = term - l_term
                        r_idx = complete_index_multisets.index(r_term)
                        lmult_indices.append([l_idx, r_idx])

            lmult_rule.append(
                (np.ones(len(lmult_indices), dtype=float), np.array(lmult_indices, dtype=int))
            )

    return lmult_rule



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Wrapper for jax.experimental.ode.odeint
"""

from typing import Callable, Optional
from scipy.integrate._ivp.ivp import OdeResult

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias import ArrayLike, requires_array_library

from .solver_utils import merge_t_args_jax, trim_t_results_jax

try:
    from jax.experimental.ode import odeint
except ImportError:
    pass


@requires_array_library("jax")
def jax_odeint(
    rhs: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
):
    """Routine for calling `jax.experimental.ode.odeint`

    Args:
        rhs: Callable of the form :math:`f(t, y)`
        t_span: Interval to solve over.
        y0: Initial state.
        t_eval: Optional list of time points at which to return the solution.
        **kwargs: Optional arguments to be passed to ``odeint``.

    Returns:
        OdeResult: Results object.
    """

    t_list = merge_t_args_jax(t_span, t_eval)

    # determine direction of integration
    t_direction = unp.sign(unp.asarray(t_list[-1] - t_list[0], dtype=complex))

    results = odeint(
        lambda y, t: rhs(unp.real(t_direction * t), y) * t_direction,
        y0=unp.asarray(y0, dtype=complex),
        t=unp.real(t_direction) * unp.asarray(t_list),
        **kwargs,
    )

    results = OdeResult(t=t_list, y=results)

    return trim_t_results_jax(results, t_eval)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
======================================
Models (:mod:`qiskit_dynamics.models`)
======================================

.. currentmodule:: qiskit_dynamics.models

This module contains classes for constructing the right-hand side of an ordinary differential
equations. In this package, a "model of a quantum system" means a description of a differential
equation used to model a physical quantum system, which in this case is either the *Schrodinger
equation*:

.. math::
    \dot{y}(t) = -i H(t)y(t),

where :math:`H(t)` is the Hamiltonian, or the *Lindblad equation*:

.. math::
    \dot{\rho}(t) = -i[H(t), \rho(t)] +
                \sum_j g_j(t) \left(L_j\rho(t)L_j^\dagger -
                \frac{1}{2}\{L_j^\dagger L_j, \rho(t)\}\right),

where the second term is called the *dissipator* term. Each :math:`L_j` is a dissipation operator
*dissipator*, and :math:`[\cdot, \cdot]` and :math:`\{\cdot, \cdot\}` are, respectively, the matrix
commutator and anti-commutator.

The classes for representing the Schrodinger and Lindblad equations are, respectively,
:class:`~qiskit_dynamics.models.HamiltonianModel` and
:class:`~qiskit_dynamics.models.LindbladModel`. Model classes primarily serve a *computational*
purpose, and expose functions for evaluating model expressions, such as :math:`t \mapsto H(t)` or
:math:`t,y \mapsto -iH(t)y` in the case of a Hamiltonian, and with similar functionality for
:class:`~qiskit_dynamics.models.LindbladModel`.

.. _Rotating frames:

Rotating frames
===============

Frame transformations are a common technique for solving time-dependent quantum differential
equations. For example, for a Hamiltonian, this corresponds to the transformation

.. math::
    H(t) \mapsto e^{iH_0t}(H(t) - H_0)e^{-iH_0t},

for a Hermitian operator :math:`H_0` called the *frame operator*.

.. note::
    The *frame operator* is commonly equivalently expressed as the corresponding anti-Hermitian
    operator under the association :math:`F = -iH_0`. This package refers to either :math:`F` or
    :math:`H_0` as the *frame operator*, with this association being understood.

Any model class can be transformed into a rotating frame by setting the
:attr:`~qiskit_dynamics.models.BaseGeneratorModel.rotating_frame` property:

.. code-block:: python

    model.rotating_frame = frame_operator

where ``frame_operator`` is a specification of either :math:`H_0` or :math:`F = -iH_0` (see the
documentation for :class:`~qiskit_dynamics.models.RotatingFrame` for valid types and behaviours).
Setting this property modifies the behaviour of the evaluation functions, e.g. a
:class:`~qiskit_dynamics.models.HamiltonianModel` will compute :math:`e^{-tF}(-iH(t) - F)e^{tF}` in
place of :math:`H(t)`. :class:`~qiskit_dynamics.models.LindbladModel` has similar behaviour.

Internally, the model classes make use of the :class:`~qiskit_dynamics.models.RotatingFrame` class,
which is instantiated when the ``rotating_frame`` property is set. This class contains helper
functions for transforming various objects into and out of the rotating frame. This class works
directly with the anti-Hermitian form :math:`F = -iH_0`, however, it can also be instantiated with a
Hermitian operator :math:`H_0` from which :math:`F` is automatically constructed.

Rotating wave approximation
===========================

The rotating wave approximation (RWA) is a transformation in which rapidly oscillating
time-dependent components, above a given cutoff frequency, are removed from a model. This
transformation is implemented in :meth:`~qiskit_dynamics.models.rotating_wave_approximation`, see
its documentation for details.

.. _model evaluation:

Controlling model evaluation: array libraries and vectorization
===============================================================

The underlying array library used by any model class can be controlled via the ``array_library``
instantiation argument. The model will store the underlying arrays using the specified library, and
use this library to evaluate the model. See the :mod:`.arraylias` submodule API documentation for a
list of ``array_library`` options. If unspecified, the model will use the general dispatching rules
of the configured aliases in :mod:`.arraylias` to determine which library to use based on how the
operators are specified.

Additionally, the :class:`.LindbladModel` class can be set to store and evaluate the Lindblad
equation in a matrix-vector vectorized format using the ``vectorized`` instatiation argument.

.. note::

    When setting a rotating frame, models internally store their operators in the basis in which the
    frame operator is diagonal. In general, sparsity of an operator is not perserved by basis
    transformations. Hence, preserving internal sparsity with rotating frames requires more
    restrictive choice of frames. For example, diagonal frame operators exactly preserve sparsity.


Model classes
=============

.. autosummary::
   :toctree: ../stubs/

   HamiltonianModel
   LindbladModel
   GeneratorModel

Model transformations
=====================

.. autosummary::
   :toctree: ../stubs/

   RotatingFrame
   rotating_wave_approximation
"""

from .rotating_frame import RotatingFrame
from .generator_model import BaseGeneratorModel, GeneratorModel
from .hamiltonian_model import HamiltonianModel
from .lindblad_model import LindbladModel
from .rotating_wave_approximation import rotating_wave_approximation



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Orthonormal basis classes.
"""

from typing import List, Optional, Callable, Tuple, Union
from itertools import product

import numpy as np

from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias.alias import _preferred_lib

from .subsystem import Subsystem
from .abstract_subsystem_operators import AbstractSubsystemOperator


class ONBasis:
    """Represents a list of orthonormal vectors."""

    def __init__(
        self,
        subsystems: List[Subsystem],
        basis_vectors: Optional[np.ndarray] = None,
        labels: Optional[List] = None,
    ):
        """An orthonormal basis specified as a 2d array whose columns are the vectors.

        Args:
            subsystems: A list of Subsystem instances. Tensor product ordering is assumed to be
                reversed.
            basis_vectors: The vectors given as a 2d array, with the vectors being the columns. If
                ``None``, defaults to the standard basis.
            labels: Labels for the basis vectors. If ``None``, defaults to integer ordering based
                on subsystem dimensions.
        """
        self._subsystems = subsystems
        subsystem_dims = [x.dim for x in subsystems]

        if basis_vectors is None:
            basis_vectors = np.eye(np.prod(subsystem_dims), dtype=complex)

        if basis_vectors.ndim != 2:
            raise ValueError("basis_vectors must be supplied as a 2d array.")

        if basis_vectors.shape[0] != np.prod(subsystem_dims):
            raise QiskitError("Vector dimension does not match subsystem dimension.")

        self._basis_vectors = basis_vectors
        self._basis_vectors_adj = self.basis_vectors.conj().transpose()

        gram_matrix = self.basis_vectors_adj @ self.basis_vectors

        if not np.allclose(gram_matrix, np.eye(len(gram_matrix))):
            raise QiskitError("Vectors are not orthonormal.")

        self._labels = labels or _default_indexing(subsystems, basis_vectors.shape[1])

    @property
    def subsystems(self):
        """Subsystems representing the tensor product space on which the basis is defined."""
        return self._subsystems

    @property
    def labels(self):
        """Basis element labels."""
        return self._labels

    @property
    def basis_vectors(self):
        """Basis vectors as the columns of an 2-d array."""
        return self._basis_vectors

    @property
    def basis_vectors_adj(self):
        """Adjoint of the basis vectors matrix."""
        return self._basis_vectors_adj

    @property
    def projection(self):
        """Matrix for the orthogonal projection onto the subspace spanned by the basis."""
        return self.basis_vectors @ self.basis_vectors_adj

    def probabilities(self, x: np.ndarray):
        """Treating x as a state vector or density matrix, compute the probabilities of observing
        the outcomes of a measurement defined by the basis vectors.

        Args:
            x: The statevector or density matrix. Which case is determined by ``x.ndim``.
        Returns:
            A 1-d array representing probabilities computed according to the Born rule.
        """
        if x.ndim == 1:
            return unp.abs(self.decompose(x)) ** 2
        elif x.ndim == 2:
            return unp.real(unp.diag(self.basis_vectors_adj @ x @ self.basis_vectors))
        else:
            raise QiskitError("ONBasis.probabilities not defined for a >2d array.")

    def decompose(self, x):
        """Return the coefficients of the projection of a vector x in the basis."""
        return self.basis_vectors_adj @ x

    def project(self, x):
        """Project a vector x onto the subspace spanned by the basis."""
        return self.projection @ x

    def subset(self, condition: Callable):
        """Get a new ONBasis consisting of a subset of this one filtered according to the condition
        function defined on the basis labels.

        Args:
            condition: A boolean-valued function on the labels of this instance of ``ONBasis``.
        Returns:
            ONBasis: An ``ONBasis`` consisting of the (vector, label) pairs in this one for which
                ``condition(label) == True``.
        """

        indices_to_include = []
        sublabels = []
        for idx, label in enumerate(self.labels):
            if condition(label):
                indices_to_include.append(idx)
                sublabels.append(label)

        indices_to_include = np.array(indices_to_include)
        return ONBasis(
            basis_vectors=self.basis_vectors[:, indices_to_include],
            subsystems=self.subsystems,
            labels=sublabels,
        )

    def __len__(self):
        return self.basis_vectors.shape[1]


class DressedBasis(ONBasis):
    """A basis with labels of the form {"index": index, "eval": eval}, where each eval is a float
    representing an eigenvalue.
    """

    def __init__(self, subsystems, basis_vectors, evals, indices: Optional[List] = None):
        """Initialize a basis where each element has an associated eval.

        The labels for the basis vectors will be dictionaries with keys "index" and "eval". Note
        that the ``ground_state`` property will always return the first basis vector.

        Args:
            subsytems: List of subsystems.
            basis_vectors: 2d array of basis vectors.
            evals: 1d array or list of eigenvalues.
            indices: Optional indexing of basis vectors (defaults to standard).
        """

        indices = indices or _default_indexing(subsystems, basis_vectors.shape[1])
        labels = [{"index": idx, "eval": eval} for idx, eval in zip(indices, evals)]

        super().__init__(subsystems=subsystems, basis_vectors=basis_vectors, labels=labels)

    @classmethod
    def from_hamiltonian(
        cls,
        hamiltonian: Union[np.ndarray, AbstractSubsystemOperator],
        subsystems: List[Subsystem],
        ordering: str = "default",
    ):
        """Build a DressedBasis instance from the eigendecomposition of a Hamiltonian, ordered in
        terms of non-decreasing eigenvalues.

        Args:
            hamiltonian: The Hamiltonian operator.
            subsystems: A list of subsystem instances.
            ordering: The ordering with which to set the basis. The value ``"default"`` returns the
                basis ordered according to the entries of the vectors with the maximum absolute
                value.
        """

        if isinstance(hamiltonian, AbstractSubsystemOperator):
            hamiltonian = hamiltonian.matrix(ordered_subsystems=subsystems)

        if ordering == "default":
            evals, evecs = _sorted_eigh(hamiltonian)
        else:
            raise QiskitError(f"Ordering type '{ordering}' not recognized.")

        return cls(subsystems=subsystems, basis_vectors=evecs, evals=evals)

    @property
    def ground_state(self):
        """Return the first basis vector."""
        return self.basis_vectors[:, 0]

    @property
    def evals(self):
        """The eigenvalues."""
        return unp.array([x["eval"] for x in self.labels])

    @property
    def computational_states(self):
        """Get subspace of all states with subsystem indices <=0."""

        def condition(label):
            return not any(x > 1 for x in label["index"])

        return self.subset(condition)

    def low_energy_states(self, cutoff_energy):
        """Return a DressedBasis object corresponding to a low energy subspace."""

        new_evals = []
        new_indices = []
        new_basis_vectors = []
        for label, vector in zip(self.labels, self.basis_vectors.transpose()):
            if label["eval"] < cutoff_energy:
                new_evals.append(label["eval"])
                new_indices.append(label["index"])
                new_basis_vectors.append(vector)

        return DressedBasis(
            subsystems=self.subsystems,
            basis_vectors=unp.array(new_basis_vectors).transpose(),
            evals=new_evals,
            indices=new_indices,
        )


def _default_indexing(subsystems, num):
    subsystem_dims = [s.dim for s in subsystems]
    subsystem_state_labels = [range(dim) for dim in subsystem_dims]
    big_prod = product(*subsystem_state_labels)
    labels = []
    for x in big_prod:
        label = list(x)
        label.reverse()
        labels.append(tuple(label))
    return labels[:num]


def _sorted_eigh(H: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Given a Hermitian operator ``H``, return the output of ``jnp.linalg.eigh``, but with
    the eigenvalues and eigenvectors sorted so that the argmax of the absolute value of the
    eigenvectors is in non-decreasing order. This also ensures that the diagonal of the matrix
    containing the eigenvectors is positive.

    The motivation for this function is when computing dressed bases, in which ``H`` is nearly
    diagonal, and therefore the eigenvectors (if there are no degeneracies) are guaranteed to be
    close to the elementary basis vectors. In such a scenario, this function ensures that the
    eigenvectors are ordered according to which elementary basis vector they are closest to, and the
    largest (in absolute value) entry of each eigenvector will be positive. If the conditions of
    this motivating example are not met, this function will still return a valid diagonalization of
    ``H``, however the ordering of the eigenvectors will not have any specific meaning.

    Args:
        H: The Hermitian operator.

    Returns:
        Tuple: The sorted eigenvalues and eigenvectors.
    """

    evals, evecs = unp.linalg.eigh(H)

    # sort based on largest entry of each column
    evecs_trans = evecs.transpose()

    if _preferred_lib(H) == "jax":
        import jax.numpy as jnp
        from jax import vmap

        max_indices_argsort = vmap(jnp.argmax)(unp.abs(evecs_trans)).argsort()
    else:
        max_indices_argsort = np.array([np.argmax(np.abs(x)) for x in evecs_trans]).argsort()

    sorted_evecs_trans = evecs_trans[max_indices_argsort]
    sorted_evals = evals[max_indices_argsort]

    # ensure all largest entries are positive
    exp_factors = unp.exp(-1j * unp.angle(unp.diag(sorted_evecs_trans)))
    sorted_evecs_trans = exp_factors[:, np.newaxis] * sorted_evecs_trans

    return sorted_evals, sorted_evecs_trans.transpose()



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Global alias instances.
"""
import functools

from types import FunctionType

from typing import Union, Callable

from scipy.sparse import spmatrix

from arraylias import numpy_alias, scipy_alias

from arraylias.exceptions import LibraryError

from qiskit import QiskitError

from .register_functions import (
    register_asarray,
    register_matmul,
    register_multiply,
    register_rmatmul,
    register_linear_combo,
    register_transpose,
    register_conjugate,
)

# global NumPy and SciPy aliases
DYNAMICS_NUMPY_ALIAS = numpy_alias()
DYNAMICS_SCIPY_ALIAS = scipy_alias()

DYNAMICS_NUMPY = DYNAMICS_NUMPY_ALIAS()
DYNAMICS_SCIPY = DYNAMICS_SCIPY_ALIAS()

# register required custom versions of functions for sparse type here
DYNAMICS_NUMPY_ALIAS.register_type(spmatrix, lib="scipy_sparse")

try:
    from jax.experimental.sparse import BCOO

    # register required custom versions of functions for BCOO type here
    DYNAMICS_NUMPY_ALIAS.register_type(BCOO, lib="jax_sparse")
except ImportError:
    pass

# register custom functions for numpy_alias
register_asarray(alias=DYNAMICS_NUMPY_ALIAS)
register_matmul(alias=DYNAMICS_NUMPY_ALIAS)
register_multiply(alias=DYNAMICS_NUMPY_ALIAS)
register_rmatmul(alias=DYNAMICS_NUMPY_ALIAS)
register_linear_combo(alias=DYNAMICS_NUMPY_ALIAS)
register_conjugate(alias=DYNAMICS_NUMPY_ALIAS)
register_transpose(alias=DYNAMICS_NUMPY_ALIAS)


ArrayLike = Union[Union[DYNAMICS_NUMPY_ALIAS.registered_types()], list]


def _isArrayLike(x: any) -> bool:
    """Return true if x is an ArrayLike object. Equivalent to isinstance(x, ArrayLike)."""
    return isinstance(x, (DYNAMICS_NUMPY_ALIAS.registered_types(), list))


def _preferred_lib(*args, **kwargs):
    """Given a list of args and kwargs with potentially mixed array types, determine the appropriate
    library to dispatch to.

    For each argument, DYNAMICS_NUMPY_ALIAS.infer_libs is called to infer the library. If all are
    "numpy", then it returns "numpy", and if any are "jax", it returns "jax".

    Args:
        *args: Positional arguments.
        **kwargs: Keyword arguments.
    Returns:
        str
    Raises:
        QiskitError: if none of the rules apply.
    """
    args = list(args) + list(kwargs.values())
    if len(args) == 1:
        libs = DYNAMICS_NUMPY_ALIAS.infer_libs(args[0])
        return libs[0] if len(libs) > 0 else "numpy"

    lib0 = _preferred_lib(args[0])
    lib1 = _preferred_lib(args[1:])

    if lib0 == "numpy" and lib1 == "numpy":
        return "numpy"
    elif lib0 == "jax_sparse" or lib1 == "jax_sparse":
        return "jax_sparse"
    elif lib0 == "jax" or lib1 == "jax":
        return "jax"
    elif lib0 == "scipy_sparse" or lib1 == "scipy_sparse":
        return "scipy_sparse"

    raise QiskitError("_preferred_lib could not resolve preferred library.")


def _numpy_multi_dispatch(*args, path, **kwargs):
    """Multiple dispatching for NumPy.

    Given *args and **kwargs, dispatch the function specified by path, to the array library
    specified by _preferred_lib.

    Args:
        *args: Positional arguments to pass to function specified by path.
        path: Path in numpy module structure.
        **kwargs: Keyword arguments to pass to function specified by path.
    Returns:
        Result of evaluating the function at path on the arguments using the preferred library.
    """
    lib = _preferred_lib(*args, **kwargs)
    return DYNAMICS_NUMPY_ALIAS(like=lib, path=path)(*args, **kwargs)


def _to_dense(x):
    """Convert an array to dense."""
    libs = DYNAMICS_NUMPY_ALIAS.infer_libs(x)
    if "scipy_sparse" in libs or "jax_sparse" in libs:
        return DYNAMICS_NUMPY.array(x.todense())
    return x


def _to_dense_list(x):
    """Convert a list of arrays to their corresponding dense version. Assumes input is either a list
    of scipy sparse, a BCOO array, or numpy/jax array.
    """
    libs = DYNAMICS_NUMPY_ALIAS.infer_libs(x)
    if "scipy_sparse" in libs:
        return DYNAMICS_NUMPY.array([op.todense() for op in x])
    elif "jax_sparse" in libs:
        return x.todense()
    return x


def requires_array_library(lib: str) -> Callable:
    """Return a function and class decorator for checking a library is available.

    If the the required library is not in the list of the registered library
    for global alias instances, any decorated function or method will raise
    an exception when called, and any decorated class will raise an exeption
    when its ``__init__`` is called.

    Args:
        lib: the library name required by class or function.

    Returns:
        Callable: A decorator that may be used to specify that a function, class,
                  or class method requires a specific library to be installed.
    """

    def decorator(obj):
        """Specify that the decorated object requires a specifc Array library."""

        def check_library(descriptor):
            if lib not in DYNAMICS_NUMPY_ALIAS.registered_libs():
                raise LibraryError(
                    f"Array library '{lib}' required by {descriptor} "
                    "is not installed. Please install the optional "
                    f"library '{lib}'."
                )

        # Decorate a function or method
        if isinstance(obj, FunctionType):

            @functools.wraps(obj)
            def decorated_func(*args, **kwargs):
                check_library(f"function {obj}")
                return obj(*args, **kwargs)

            return decorated_func

        # Decorate a class
        elif isinstance(obj, type):
            obj_init = obj.__init__

            @functools.wraps(obj_init)
            def decorated_init(self, *args, **kwargs):
                check_library(f"class {obj}")
                obj_init(self, *args, **kwargs)

            obj.__init__ = decorated_init
            return obj

        else:
            raise ValueError(f"Cannot decorate object {obj} that is not a class or function.")

    return decorator
